/**
 * Filter object with filter function `fn`.
 *
 * @param obj Object to filter.
 * @param fn Filter function receiving [key, value].
 * @returns A new object filtered.
 */ const filter = (obj, fn)=>Object.fromEntries(Object.entries(obj).filter(fn));
/**
 * Pick properties from an object into a new object.
 *
 * @param obj Object to pick from.
 * @param props Properties to pick.
 * @returns A new object with the picked properties.
 */ const pick = (obj, props)=>props.reduce((p, n)=>{
        if (n in obj) p[n] = obj[n];
        return p;
    }, {});
/**
 * Create a new object with certain properties omitted.
 *
 * @param obj Object to omit from.
 * @param props Properties to omit.
 * @returns A new object with the properties omitted.
 */ const omit = (obj, props)=>filter(obj, includesKey(props, true));
/**
 * @private
 */ const includesKey = (props, invert = false)=>([key])=>invert ^ props.includes(key);

const bool = Symbol.for('to-fluent-bool');
/**
 * Convert a function with a settings object to fluent API.
 *
 * ```ts
 * import { toFluent, bool } from 'to-fluent'
 *
 * const cb = toFluent(
 *   class {
 *     foo = bool // indicate boolean but initially omitted
 *     bar?: string // optional
 *     zoo = 123 // a default
 *   },
 *   settings => () => settings
 * )
 *
 * expect(cb()).toEqual({ zoo: 123 })
 * expect(cb.foo()).toEqual({ foo: true, zoo: 123 })
 * expect(cb.not.foo()).toEqual({ foo: false, zoo: 123 })
 * expect(cb.bar('hello')()).toEqual({ bar: 'hello', zoo: 132 })
 * expect(cb.foo.bar('hello').zoo(456)())
 *   .toEqual({ foo: true, bar: 'hello', zoo: 456 })
 * ```
 */ const toFluent = (Schema, cb)=>{
    const bools = Object.entries(new Schema()).filter(([, x])=>x === bool || typeof x === 'boolean');
    const flags = bools.map(([key])=>key);
    const omitted = bools.filter(([, x])=>x === bool).map(([key])=>key);
    const settings = omit(new Schema(), omitted);
    let not = false;
    const bind = (settings)=>new Proxy(cb, {
            get (_, key, receiver) {
                if (key === 'not') {
                    not = true;
                    return receiver;
                } else if (flags.includes(key)) {
                    const value = not ? false : true;
                    not = false;
                    return bind({
                        ...settings,
                        [key]: value
                    });
                } else {
                    return (value)=>{
                        return bind({
                            ...settings,
                            [key]: value
                        });
                    };
                }
            },
            apply: (_, self, args)=>cb.call(self, {
                    ...settings
                }).apply(self, args)
        });
    return bind(settings);
};

const entries = (obj)=>Object.entries(obj);
const fromEntries = (entries)=>Object.fromEntries(entries);
const cheapRandomId = ()=>(Math.random() * 10e7 | 0).toString(36);
const accessors = (target, source, fn, filter = ()=>true)=>{
    const prevDesc = new Map();
    Object.defineProperties(target, Object.fromEntries(entries(source).filter(([key])=>typeof key === 'string').filter(([key, value])=>filter(key, value)).map(([key, value])=>{
        const next = fn(key, value);
        const prev = Object.getOwnPropertyDescriptor(target, key);
        prevDesc.set(key, prev);
        if (prev && prev.get && prev.set) {
            const { get , set  } = next;
            next.get = ()=>{
                // run both getters if any, for side-effects to run
                // the correct behavior here? maybe pass the previous
                // getter as a param to the next one and let them decide?
                // but it works.
                const top = get?.();
                const below = prev.get();
                return below ?? top;
            };
            next.set = (v)=>{
                prev.set?.(v);
                set(prev.get?.() ?? v);
            };
        }
        //!? 'create accessor', key, prev, next
        return [
            key,
            {
                configurable: true,
                enumerable: true,
                ...next
            }
        ];
    })));
    return ()=>{
        for (const [key, desc] of prevDesc){
            if (desc == null) {
                delete target[key];
            } else {
                // const newValue = (target as any)[key]
                Object.defineProperty(target, key, desc);
            }
        }
    };
};
const kebab = (s)=>s.replace(/[a-z](?=[A-Z])|[A-Z]+(?=[A-Z][a-z])/g, '$&-').toLowerCase();
const styleToCss = (style)=>{
    let css = '';
    for(const key in style)css += style[key] ? kebab(key) + ':' + style[key] + ';' : '';
    return css;
};
const removeFromArray = (arr, el, quiet = false)=>{
    const index = arr.indexOf(el);
    if (!~index) {
        if (quiet) return;
        throw new ReferenceError('Item to be removed does not exist in the array.');
    }
    return arr.splice(index, 1);
};
const shallowEqual = (a, b)=>[
        [
            a,
            b
        ],
        [
            b,
            a
        ]
    ].every(([a, b])=>entries(a).every(([key, value])=>key in b && b[key] == value));
const getOwnProperty = (object, name)=>Object.getOwnPropertyDescriptor(object, name)?.value;
const defineProperty = toFluent(class {
    configurable = false;
    enumerable = false;
    writable = bool;
}, (descriptor)=>(object, name, value)=>Object.defineProperty(object, name, value == null ? descriptor : {
            ...descriptor,
            value
        }));
const filterMap = (array, fn)=>array.map(fn).filter(Boolean);

const isEvent = (name)=>name.charAt(0) === 'o' && name.charAt(1) === 'n' && !name.endsWith('ref');
const toEvent = (name)=>name.slice(2);
const createProp = (doc = html, el, _type, name, value, attrs)=>{
    // all the Any's below are on purpose
    // all the cases should be taken care of,
    // but since this is user input they are allowed
    // to do something wrong and we want to raise an
    // special cases
    switch(name){
        case 'children':
            return;
        case 'style':
            // if we createAttribute and set .value then that
            // triggers the css parser and we can't compare if
            // doing it this way retains the exact string and
            value = value?.valueOf();
            if (typeof value === 'object') value = styleToCss(value);
            if (value) {
                el.setAttribute('style', value);
                // we know there is an attribute node because we
                attrs.style = el.getAttributeNode('style');
            }
            return;
    }
    //
    // create prop
    // we used to normalize the attribute name
    // but now we don't. because there is a high
    // normalizes attribute names we leave it here
    const attr = name // toAttr[name] || name
    ;
    value = value?.valueOf();
    switch(typeof value){
        case 'string':
        case 'number':
            if (doc === svg || !(name in el)) {
                el.setAttribute(attr, value);
                attrs[attr] = el.getAttributeNode(attr);
                return;
            }
            break;
        case 'function':
            if (isEvent(attr)) {
                //!? 'create event listener', attr, value
                el.addEventListener(toEvent(attr), value, value);
            } else {
                el.setAttribute(attr, '');
                attrs[attr] = el.getAttributeNode(attr);
                el[name] = value;
            }
            return;
    }
    if (value == null) {
        el.removeAttribute(attr);
        delete attrs[attr];
        if (el[name] != null) {
            delete el[name];
        }
        return;
    }
    el[name] = value;
    if (el.hasAttribute(attr)) attrs[attr] = el.getAttributeNode(attr);
};
const propCache = new WeakMap();
const createProps = (doc, el, type, props = {}, attrs = {}, cacheRef = el)=>{
    for(const name in props)createProp(doc, el, type, name, props[name], attrs);
    propCache.set(cacheRef, {
        props,
        attrs
    });
};
const updateProps = (doc, el, type, next = {}, cacheRef = el)=>{
    if (!propCache.has(cacheRef)) return next && createProps(doc, el, type, next, void 0, cacheRef);
    const c = propCache.get(cacheRef);
    const { attrs , props  } = c;
    if (!next) {
        //!? 'disconnected, removing all attrs and props', cacheRef
        for(const name in attrs){
            el.removeAttributeNode(attrs[name]);
        }
        for(const name1 in props){
            if (isEvent(name1)) {
                //!? 'removed event listener', name
                el.removeEventListener(toEvent(name1), props[name1], props[name1]);
            } else {
                delete el[name1];
            }
        }
        propCache.delete(cacheRef);
        return;
    }
    let prev;
    let value;
    out: for(const name2 in props){
        if (!(name2 in next)) {
            //!? 'removed prop', name
            delete el[name2];
            continue;
        }
        value = next[name2];
        switch(name2){
            case 'children':
                continue out;
        }
        value = value?.valueOf()
        ;
        prev = props[name2];
        if (prev !== value) {
            if (typeof value === 'function') {
                if (prev.fn && value.fn) {
                    //!? 'updated fn'
                    value = prev.update(value.fn, value.options);
                    if (value === prev) {
                        next[name2] = value;
                    }
                }
                if (prev !== value || cacheRef !== el) {
                    let attr = name2 // toAttr[name] || name
                    ;
                    if (isEvent(attr)) {
                        attr = toEvent(attr);
                        //!? 'updated event listener', attr, prev, value
                        el.removeEventListener(attr, prev, prev);
                        el.addEventListener(attr, value, value);
                    } else {
                        props[attr] = el[attr] = value;
                    }
                }
            } else if (!(name2 in attrs)) {
                if (value == null) {
                    //!? 'removed prop', name
                    if (el[name2] != null) {
                        el[name2] = value;
                    }
                } else {
                    el[name2] = value;
                }
            }
        }
    }
    for(const name3 in attrs){
        if (!(name3 in next) || next[name3] === false || next[name3] == null) {
            //!? 'removed attr', name
            el.removeAttribute(name3)
            ;
            delete attrs[name3];
            continue;
        }
        value = next[name3]?.valueOf();
        switch(name3){
            case 'style':
                if (typeof value === 'object') value = styleToCss(value) //|| null
                ;
                break;
        }
        // updated value
        if (props[name3] !== value && typeof value !== 'function') {
            if (value == null) {
                if (attrs[name3].value != null) {
                    //!? 'removed attr', name
                    el.removeAttribute(name3);
                    delete attrs[name3];
                }
            } else {
                //!? 'updated attr', name, value
                attrs[name3].value = value;
            }
        }
    }
    for(const name4 in next){
        if (!(name4 in attrs) && !(name4 in props)) createProp(doc, el, type, name4, next[name4], attrs);
    }
    c.props = next;
};

const anchor = new Comment();
const Fragment = Symbol();
const jsx = (kind, props, key)=>kind === Fragment ? props.children : {
        kind,
        props,
        key
    };
const jsxs = jsx;
let hook;
const createHook = ()=>function current(fn = current.fn) {
        const prev = hook;
        hook = current;
        hook.fn = fn;
        fn();
        hook = prev;
    };
let Chunk = class Chunk extends Array {
    firstChild;
    dom = [];
    get last() {
        return this.dom.at(-1);
    }
    get nextSibling() {
        return this.last?.nextSibling;
    }
    //   this.unshift(x)
    appendChild(x) {
        this.push(x);
    }
    after(x) {
        this.last?.after(x);
    }
    save() {
        this.dom = [
            ...this
        ];
    }
    remove() {
        this.dom.forEach((el)=>{
            el.onunref?.();
            el.remove();
        });
        this.splice(0);
    }
    removeChild(x) {
        const i = this.indexOf(x);
        ~i && this.splice(i, 1);
    // i = this.dom.indexOf(x)
    // ~i && this.dom.splice(i, 1)
    }
};
const { TEXT_NODE , COMMENT_NODE  } = document;
const html = document.createElement.bind(document);
const svg = document.createElementNS.bind(document, 'http://www.w3.org/2000/svg');
const forceArray = (x, withNull)=>Array.isArray(x) ? withNull && !x.length ? [
        null
    ] : x : x == null && !withNull ? [] : [
        x
    ];
const flatDom = (arr, res = [])=>{
    for (const el of arr){
        if (el.dom) res.push(...flatDom(el.dom));
        else res.push(el);
    }
    return res;
};
const prevs = new WeakMap();
function render(n, el = document.createDocumentFragment(), doc = html, withNull = false) {
    reconcile(el, forceArray(n, withNull), prevs.get(el), doc);
    return el;
}
const reconcile = (parent, nk, pk, doc)=>{
    if (pk?.running) {
        console.warn('attempt to reconcile lane which is reconciling');
        return;
    }
    if (pk === nk) nk = [
        ...nk
    ];
    for (const [i, n] of nk.entries()){
        nk[i] = n?.valueOf?.();
    }
    prevs.set(parent, nk);
    nk.running = true;
    nk.dom = Array(nk.length);
    nk.keyed = new Map();
    nk.mapped = new Map();
    if (Array.isArray(pk)) {
        const keep = new Set();
        for(let i1 = 0, n1, el, p, pel, k, pi; i1 < nk.length; i1++){
            n1 = nk[pi = i1];
            k = n1?.key;
            if (k != null) {
                nk.keyed.set(k, i1);
                pi = pk.keyed.get(k) ?? -1;
            }
            p = pk[pi];
            pel = pk.dom[pi];
            nk.dom[i1] = el = create(doc, n1, p, pel);
            if (el === pel) keep.add(pel);
        }
        for (const pel1 of pk.dom){
            if (!keep.has(pel1)) {
                pel1.onunref?.();
                if (parent instanceof Chunk) parent.removeChild(pel1);
                pel1.remove();
                pk.mapped.get(pel1)?.hook?.onremove?.();
            }
        }
    } else {
        for(let i2 = 0, n2, el1, k1; i2 < nk.length; i2++){
            n2 = nk[i2];
            k1 = n2?.key;
            if (k1 != null) nk.keyed.set(k1, i2);
            nk.dom[i2] = el1 = create(doc, n2);
            nk.mapped.set(el1, n2);
        }
    }
    nk.flatDom = flatDom(nk.dom);
    if (pk?.flatDom) diff(parent, nk.flatDom, pk.flatDom);
    else nk.flatDom.forEach((el)=>parent.appendChild(el));
    nk.running = false;
};
// scheduling to prevent triggering ref effects before this render finishes
// TODO: has to be put in a proper queue instead of relying on the microtask
const mount = (el)=>{
    queueMicrotask(()=>{
        if (el?.ref && el !== el.ref.current) {
            el.ref.current = el;
        }
        if (el?.onref) {
            el.onunref = el.onref(el);
        }
    });
    return el;
};
const diff = (parent, n, p, i = 0, len = n.length, el, last)=>{
    if (parent instanceof Chunk) {
        // TODO: optimize this
        for(; i < len; i++){
            el = n[i];
            if (i < parent.length) {
                if (p[i] === el) continue;
                parent[i] = el;
            } else {
                parent.push(el);
            }
        }
        let d = parent.length - len;
        while(d--)parent.pop();
    } else {
        for(; i < len; i++){
            el = n[i];
            if (p[i] === el) last = el;
            else if (!i) parent.insertBefore(last = el, parent.firstChild);
            else last.after(last = el);
        }
    }
};
const create = (doc, n, p, pel)=>{
    let el;
    switch(typeof n){
        case 'string':
        case 'number':
            if (pel?.nodeType === TEXT_NODE) {
                if (p != n) pel.nodeValue = n;
                return pel;
            }
            el = new Text(n);
            return el;
        case 'object':
            if (!n) break;
            if (Array.isArray(n)) {
                if (pel && Array.isArray(p)) el = pel;
                else el = new Chunk();
                if (!n.length) n.push(null);
                reconcile(el, n, p, doc);
                el.save();
            } else if (typeof n.kind === 'string') {
                // maybe switch to svg namespace
                if (n.kind === 'svg') doc = svg;
                if (n.props.ref?.current && n.props.ref.current.tagName.toLowerCase() === n.kind && (el = n.props.ref.current) || // if the previous element is of the same kind, use that
                pel && p?.kind === n.kind && (el = pel)) {
                    updateProps(doc, el, n.kind, n.props);
                } else {
                    // otherwise create an element and props
                    el = doc(n.kind, 'is' in n.props ? {
                        is: n.props.is
                    } : void 0);
                    createProps(doc, el, n.kind, n.props);
                }
                if (n.kind === 'foreignObject') doc = html;
                // render children
                render(n.props.children, el, doc);
                mount(el);
            } else {
                let initial = true;
                if (!((el = pel) && (n.hook = p?.hook))) {
                    el = new Chunk();
                    n.hook = createHook();
                }
                let prevDom;
                let nextDom;
                n.hook(()=>{
                    let next;
                    if (!initial && !(next = el.nextSibling)) el.after(next = anchor);
                    render(n.kind(n.props), el, doc, true);
                    el.save();
                    if (!initial && next) {
                        nextDom = flatDom(el);
                        if (prevDom?.length > 0) {
                            for(let i = 0, e; i < nextDom.length; i++){
                                e = nextDom[i];
                                if (prevDom[i] !== e) next.before(e);
                            }
                        } else {
                            for (const e1 of nextDom)next.before(e1);
                        }
                        prevDom = nextDom;
                        next === anchor && next.remove();
                    } else {
                        initial = false;
                    }
                });
            }
            return el;
    }
    if (pel?.nodeType === COMMENT_NODE) return pel;
    el = new Comment();
    return el;
};

/**
 * Creates a proxy that calls `cb` for every property accessed and returns its result.
 *
 * @param cb Called with the key of the property being accessed.
 */ const Getter = (cb)=>new Proxy({}, {
        get: (_, key)=>cb(key)
    });
const FluentCaptureSymbol = Symbol();
const FluentCapture = ()=>{
    const fn = ()=>{};
    const results = [];
    const proxy = new Proxy(fn, {
        get: (_, key, receiver)=>{
            if (key === FluentCaptureSymbol) return true;
            if (key === '_results') return results;
            results.push([
                'get',
                key
            ]);
            return receiver;
        },
        apply: (_, __, args)=>{
            results.push([
                'apply',
                args
            ]);
            return proxy;
        }
    });
    return proxy;
};
const applyFluentCapture = (capture, target)=>{
    let res = target;
    for (const [op, args] of capture){
        if (op === 'get') {
            res = res[args];
        } else if (op === 'apply') {
            res = res(...args);
        }
    }
    return res;
};

const argsOf = (x)=>{
    const d = x.indexOf('{') + 1;
    if (!d || x.slice(0, d).includes(')')) return [];
    x = x.slice(d);
    let match;
    do {
        match = false;
        x = x.replace(/`((\\`)?[^`])+?`|'((\\')?[^'])+?'|"((\\")?[^"])+?"|\([^{[(]*?\)|\{[^{[(]*?\}|\[[^{[(]*?\]/g, ()=>{
            match = true;
            return '';
        });
    }while (match)
    return x.split('}')[0].split(',').map((x)=>x.split(/[^\w\s$]+/g)[0].trim());
};
/**
 * Extract destructured argument names from `fn`'s first argument and return a `Set` of `keys`.
 *
 * ```ts
 * const fn = ({ foo, bar }) => void 0
 * argtor(fn) // => Set(2) {'foo', 'bar'}
 * ```
 *
 * @param fn The function to extract argument keys from.
 * @param keys A keys `Set` to fill in with the argument names.
 * @returns The keys `Set` with the argument names.
 */ const argtor = (fn, keys = new Set())=>(argsOf('' + fn).forEach((x)=>!x || keys.add(x)), keys);

class AbortOptions {
    throw = bool;
    latest = bool;
    timeout;
}
const abort = toFluent(AbortOptions, (options)=>(fn)=>{
        let ctrl;
        return async function(...args) {
            if (options.latest && ctrl) ctrl.abort();
            ctrl = new AbortController();
            if (options.timeout != null) {
                const timeoutError = new Error('Timed out');
                try {
                    return await Promise.race([
                        new Promise((_, reject)=>setTimeout(reject, options.timeout, timeoutError)),
                        fn(ctrl.signal).apply(this, args),
                    ]);
                } catch (error) {
                    if (error === timeoutError) {
                        ctrl.abort();
                        if (options.throw) throw error;
                    } else {
                        if (options.throw) throw error;
                    }
                }
            } else {
                return fn(ctrl.signal).apply(this, args);
            }
        };
    });

const MouseButton = {
    Left: 1,
    Right: 2,
    Middle: 4
};

class DispatchOptions {
    bubbles = bool;
    cancelable = bool;
    composed = bool;
}
const dispatchEvent = (el, nameOrEvent, detail, init, options)=>el.dispatchEvent(nameOrEvent instanceof Event ? nameOrEvent : new CustomEvent(nameOrEvent, {
        detail,
        ...init,
        ...options
    }));
const dispatch = toFluent(DispatchOptions, (options)=>(el, nameOrEvent, detail, init)=>dispatchEvent(el, nameOrEvent, detail, init, options));
const dispatchBind = (el)=>toFluent(DispatchOptions, (options)=>(nameOrEvent, detail, init)=>dispatchEvent(el, nameOrEvent, detail, init, options));

const Task = (fn, self, args)=>{
    let resolve;
    let reject;
    const promise = new Promise((rs, rj)=>{
        resolve = rs;
        reject = rj;
    });
    return {
        promise,
        resolve,
        reject,
        fn,
        self,
        args
    };
};
const runTask = (task, res)=>(task.resolve(res = task.fn.apply(task.self, task.args)), res);
const groupTasks = (other, tasks)=>tasks.forEach((t)=>other.promise.then(t.resolve));

class QueueOptions {
    first = bool;
    last = bool;
    next = bool;
    raf = bool;
    task = bool;
    time = bool;
    atomic = bool;
    debounce;
    throttle;
}
const wrapQueue = (options = {})=>(fn)=>{
        const initialOptions = {
            ...options
        };
        const queued = [];
        let queueFn;
        let id;
        let last;
        let runs = false;
        if (options.raf) queueFn = requestAnimationFrame;
        else if (options.task) queueFn = queueMicrotask;
        else if (options.time) queueFn = setTimeout;
        else if (options.throttle != null) {
            queueFn = (fn)=>{
                setTimeout(fn, options.throttle);
            };
        } else if (options.debounce != null) {
            queueFn = (fn)=>{
                clearTimeout(id);
                id = setTimeout(fn, options.debounce);
            };
        } else if (options.atomic) queueFn = (fn)=>fn();
        else {
            // No queue function provided, return identity.
            // This is used when extending this in `event`.
            return fn;
        }
        if (options.first == null && options.last == null) {
            if (options.throttle != null) options.first = true;
            else if (options.debounce != null) options.last = true;
            else options.next = true;
        }
        const cb = ()=>{
            let task;
            if (queued.length) {
                if (options.atomic) {
                    task = queued.shift();
                    runTask(task).catch((_error)=>{
                    }).finally(()=>queueFn(cb));
                    return;
                }
                if (options.last) {
                    if (options.next) {
                        const left = queued.splice(0, queued.length - 1);
                        task = left.pop() ?? queued.pop();
                        groupTasks(task, left);
                        last = runTask(task);
                        if (queued.length) {
                            queueFn(cb);
                            return;
                        }
                    } else {
                        task = queued.pop();
                        groupTasks(task, queued.splice(0));
                        last = runTask(task);
                    }
                } else if (options.next) {
                    task = queued.shift();
                    groupTasks(task, queued.splice(0, queued.length - 1));
                    queueFn(cb);
                    last = runTask(task);
                    return;
                } else {
                    task = Task();
                    groupTasks(task, queued.splice(0));
                    task.resolve(last);
                }
            }
            runs = false;
        };
        function wrapped(...args) {
            //!? 'wrap called'
            const task = Task(fn, this, args);
            if (!runs && options.first) {
                runs = true;
                last = runTask(task);
                queueFn(cb);
                return task.promise;
            }
            queued.push(task);
            if (!runs || options.debounce) {
                runs = true;
                queueFn(cb);
            }
            return task.promise;
        }
        wrapped.fn = fn;
        wrapped.options = initialOptions;
        wrapped.update = (newFn, newOptions)=>{
            //!? 'updating fn'
            if (!shallowEqual(initialOptions, newOptions)) {
                //!? 'new options', initialOptions, newOptions
                return newFn;
            }
            fn = newFn;
            return wrapped;
        };
        return wrapped;
    };
/**
 * Decorate function `fn` with a queue function.
 *
 * The decorated function will returns a `Promise` that resolves with its result.
 * All calls will be resolved with latest result at any given quantum.
 *
 * Options:
 * - `first` => Run only first, then debounce.
 * - `last` => Run last (default behavior when using `debounce`).
 * - `next` => Run final subsequent call on next quantum (default behavior when nothing is set and not a `debounce`).
 *
 * - `raf` => Queue with `requestAnimationFrame`.
 * - `task` => Queue with `queueMicrotask`.
 * - `time` => Queue with `setTimeout`.
 *
 * - `debounce(ms)` => Debounce at specified `ms`.
 * - `throttle(ms)` => Throttle at specified `ms`.
 *
 * ```ts
 * fn = (x: number) => console.log(x)
 *
 * const cb = queue.task((x: number) => count += x)
 * cb(1) // passes (after task)
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // passes
 *
 * const cb = queue.task.last((x: number) => count += x)
 * cb(1) // discarded
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // passes
 *
 * const cb = queue.task.first((x: number) => count += x)
 * cb(1) // passes (before task)
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // discarded
 *
 * const cb = queue.task.first.last((x: number) => count += x)
 * cb(1) // passes (before task)
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // passes
 *
 * const cb = queue.task.first.last.next((x: number) => count += x)
 * cb(1) // passes (before task)
 * cb(2) // discarded
 * cb(3) // passes (after task)
 * cb(4) // passes (next task)
 *
 * const cb = queue.task.last.next((x: number) => count += x)
 * cb(1) // discarded
 * cb(2) // discarded
 * cb(3) // passes (after task)
 * cb(4) // passes (next task)
 * ```
 */ const queue = toFluent(QueueOptions, wrapQueue);

class EventOptions extends QueueOptions {
    prevent = bool;
    stop = bool;
    immediate = bool;
    capture = bool;
    once = bool;
    passive = bool;
}
const wrapEvent = (options = {})=>(fn = ()=>{})=>wrapQueue(options)(options.prevent || options.stop || options.immediate || options.capture != null || options.once != null || options.passive != null ? Object.assign(function(e) {
            if (options.prevent) e.preventDefault();
            if (options.stop) {
                options.immediate ? e.stopImmediatePropagation() : e.stopPropagation();
            }
            return fn.call(this, e);
        }, options) : fn);
/**
 * Decorates event handler `fn`.
 *
 * Options:
 * - `prevent` => `event.preventDefault()`
 * - `stop` => `event.stopPropagation()`
 * - `stop.immediate` => `event.stopImmediatePropagation()`
 *
 * Listener options:
 * - `capture` => If specified, indicates that events of this type will be dispatched to the registered listener before being dispatched to any `EventTarget` beneath it in the DOM tree.
 * - `once` => If specified, indicates that the listener should be invoked at most once after being added. The listener would be automatically removed when invoked.
 * - `passive` => Indicates that the function specified by listener will never call `preventDefault()`. If a passive listener does call `preventDefault()`, the user agent will do nothing other than generate a console warning. If not specified, defaults to `not.passive` – except that in browsers other than Safari and Internet Explorer, defaults to `passive` for the `wheel`, `mousewheel`, `touchstart` and `touchmove` events.
 *
 * Queue options:
 * - `first` => Run only first, then debounce.
 * - `last` => Run last (default behavior when using `debounce`).
 * - `next` => Run final subsequent call on next quantum (default behavior when nothing is set and not a `debounce`).
 *
 * - `raf` => Queue with `requestAnimationFrame`.
 * - `task` => Queue with `queueMicrotask`.
 * - `time` => Queue with `setTimeout`.
 *
 * - `debounce(ms)` => Debounce at specified `ms`.
 * - `throttle(ms)` => Throttle at specified `ms`.
 *
 * ```ts
 * btn.onclick = event(fn)
 * btn.onclick = event.prevent(fn)
 * btn.onclick = event.prevent.stop(fn)
 * btn.onclick = event.stop.immediate(fn)
 *
 * // examples with queue
 * btn.onclick = event.stop.raf(fn)
 * btn.onclick = event.prevent.throttle(50)(fn)
 * btn.onclick = event.debounce(100)(fn)
 * ```
 */ const event = toFluent(EventOptions, wrapEvent);

function chain(first, ...rest) {
    if (Array.isArray(first)) {
        [first, ...rest] = first.flat().filter(Boolean);
        if (first == null) return;
    }
    return (...args)=>{
        first(...args);
        for (const fn of rest)fn(...args);
    };
}
const onAll = (target, listener, ...args)=>{
    const targetOwnDispatch = target.dispatchEvent;
    const events = new Set();
    target.dispatchEvent = function(event) {
        if (!events.has(event.type)) {
            target.addEventListener(event.type, listener, ...args);
            events.add(event.type);
        }
        return targetOwnDispatch.call(this, event);
    };
    return ()=>{
        for (const eventType of events){
            target.removeEventListener(eventType, listener, ...args);
        }
        events.clear();
        target.dispatchEvent = targetOwnDispatch;
    };
};

class OnOptions extends EventOptions {
    once = bool;
    passive = bool;
    capture = bool;
}
const onEvent = (el, type, listener, options)=>{
    el.addEventListener(type, listener, options);
    return ()=>el.removeEventListener(type, listener, options);
};
const onEventFluent = (el, key)=>toFluent(OnOptions, (options)=>(listener = ()=>{})=>onEvent(el, key, wrapEvent(options)(listener), options));
function on(el, key) {
    return key != null ? onEventFluent(el, key) : Getter((key)=>onEventFluent(el, key));
}

const NO_DEPS = Symbol();
const REDUCER = Symbol();
class EffectOptions extends QueueOptions {
    cb;
    keys;
    once = bool;
    lock = bool;
    isFulfillReducer = bool;
}
class Fx {
    remove;
    target;
    values = {};
    pass;
    constructor(fx){
        Object.assign(this, fx);
    }
}
let GlobalLock;
class ContextClass {
    entries;
    mem;
    keys;
    effects;
    triggered;
    active;
    reducer;
    $;
    scheduled;
    listeners;
    static attach(target, extension = {}) {
        if (target.context instanceof ContextClass) {
            // TODO: what should be the behavior here?
            return;
        }
        target.context = new ContextClass(target);
        // We are using mutate() to extend $ because at this point there are reactive effects
        // that will trigger and depend on the $ object to be fulfilled, along with its extension.
        // in the DOM context, the extension contains DOM initializers that don't exist in Workers
        target.context.mutate(()=>{
            target.$ = Object.assign(target.context.$, extension);
        });
    }
    constructor(target){
        this.target = target;
        this.keys = new Set();
        this.triggered = new Set();
        this.$ = {};
        this.scheduled = new Set();
        this.listeners = new Map();
        this.maybeAddChangeListener = (key, value)=>{
            if (typeof value !== 'object' || !(value instanceof EventTarget) || typeof Element !== 'undefined' && value instanceof Element) {
                return;
            }
            const [prev, off] = this.listeners.get(key) ?? [];
            if (prev === value) return;
            //!? 'change listener', key
            off?.();
            const ownEvents = new WeakSet();
            const runKeyEffects = ()=>{
                // clearing values so they retrigger for same reference
                this.clearKey(key)
                ;
                this.target[key] = value;
                // and then manually
                this.runEffects(key);
            };
            this.listeners.set(key, [
                value,
                onAll(value, (event)=>{
                    if (event.type === 'change') runKeyEffects();
                    if (ownEvents.has(event)) return;
                    // TODO: better copy event function
                    const newEvent = new event.constructor(event.type, {
                        ...event,
                        detail: event.detail,
                        composed: true,
                        bubbles: true
                    });
                    ownEvents.add(newEvent);
                    if (typeof Element === 'undefined' || !(this.target instanceof Element)) {
                        console.error(this.target);
                        throw new Error('Target is not an EventTarget');
                    }
                    this.target.dispatchEvent(newEvent);
                }),
            ]);
            return true;
        };
        this.cleanup = ()=>{
            if (this.effects) {
                for (const fns of this.effects.values())fns.forEach((f)=>f.dispose?.());
            }
            this.effects = new Map([
                [
                    NO_DEPS,
                    []
                ],
                ...this.entries.map(([key])=>[
                        key,
                        []
                    ]),
            ]);
        };
        this.register = (f)=>{
            f.keys = f.options?.keys ?? argtor(f.fn);
            f.fn = wrapQueue(f.options)(f.fn);
            f.keys.forEach((key)=>{
                const fx = this.effects.get(key);
                if (!fx) {
                    console.warn('No effects for key:', key);
                    return;
                }
                fx.push(f);
            });
            !f.keys.size && this.effects.get(NO_DEPS).push(f);
            f.remove = async ()=>{
                if (f.keys.size) {
                    f.keys.forEach((key)=>{
                        const fx = this.effects.get(key);
                        if (!fx) {
                            console.warn('No effects for key:', key);
                            return;
                        }
                        removeFromArray(fx, f, true);
                    });
                } else removeFromArray(this.effects.get(NO_DEPS), f);
                if (this.triggered.has(f)) this.triggered.delete(f);
                f.dispose?.();
            };
            return this.run(f);
        };
        this.flush = ()=>{
            const current = [
                ...this.triggered
            ];
            this.triggered.clear();
            current.forEach(this.run);
        };
        this.run = (f)=>{
            if (f.pass && f.options?.once) return;
            if (this.active || GlobalLock) {
                if (f !== this.active) this.triggered.add(f);
                GlobalLock?.add(this);
                return;
            }
            this.active = f;
            if (this.update(f)) {
                f.dispose?.();
                if (!f.pass) {
                    f.values = {};
                    return this.complete();
                }
            }
            return this.finalize();
        };
        this.complete = (result)=>{
            const f = this.active;
            if (!f.pass) {
                this.active = null;
                this.flush();
                return;
            }
            if (f.cb?.(result) === false) {
                this.active = null;
                this.flush();
                return;
            }
            if (f.target != null && !f.options?.isFulfillReducer) {
                this.target[f.target] = result;
                this.active = null;
                this.flush();
                return true;
            } else {
                if (typeof result === 'function') {
                    f.dispose = ()=>{
                        result();
                        f.dispose = null;
                    };
                }
                this.active = null;
                this.flush();
            }
        };
        this.reduce = toFluent(EffectOptions, (options)=>(fn, initial)=>{
                this.reducer = new Fx({
                    fn,
                    options,
                    initial
                });
                return REDUCER;
            });
        this.fulfill = toFluent(EffectOptions, (options)=>(fn, initial)=>{
                options.isFulfillReducer = true;
                this.reducer = new Fx({
                    fn,
                    options,
                    initial
                });
                return REDUCER;
            });
        this.effect = toFluent(EffectOptions, (options)=>(fn, cb)=>{
                const f = new Fx({
                    fn,
                    cb: cb ?? options.cb,
                    options
                });
                this.register(f);
                return f.remove;
            });
        this.mutate = (fn)=>{
            const f = new Fx({
                fn,
                options: {
                    lock: true
                }
            });
            f.keys = new Set();
            f.values = {
                ...this.target
            } //as Deps<T>
            ;
            return this.run(f);
        };
        this.mutating = (fn)=>()=>{
                this.mutate(fn);
            };
        this.use = (fn)=>fn(this.target);
        this.using = (fn)=>()=>fn(this.target);
        this.function = (fn)=>{
            const { withLock , target  } = this;
            function wrapped(...args) {
                let result;
                withLock(()=>{
                    const cb = fn(target);
                    result = cb.apply(this, args);
                });
                return result;
            }
            return wrapped;
        };
        this.with = (fn)=>{
            const { target  } = this;
            function wrapped(...args) {
                const cb = fn(target);
                return cb.apply(this, args);
            }
            return wrapped;
        };
        this.callback = toFluent(EffectOptions, (options)=>(fn)=>{
                const { mutate  } = this;
                const qfn = wrapQueue(options)(function wrapped(...args) {
                    mutate((ctx)=>{
                        const cb = fn(ctx);
                        cb.apply(this, args);
                    });
                });
                return qfn;
            });
        this.atomic = (fn)=>this.callback(()=>fn);
        this.query = (sel)=>this.reduce(({ root  })=>root.querySelector(sel));
        this.when = (condition, cb)=>function(...args) {
                if (condition) cb.apply(this, args);
            };
        this.withGlobalLock = (fn)=>{
            GlobalLock = new Set();
            try {
                fn();
            } catch (error) {
                console.warn(error);
            }
            const items = [
                ...GlobalLock
            ];
            GlobalLock = null;
            for (const item of items){
                item.flush();
            }
        };
        this.withLock = (fn)=>{
            this.active = fn;
            try {
                fn();
            } catch (error) {
                console.warn(error);
            }
            this.active = null;
            this.flush();
        };
        this.copyMethods();
        this.createMemory();
        this.createAccessors();
        this.cleanup();
        for (const item of this.scheduled){
            if (item.value) {
                const { key , value: [method, args]  } = item;
                this.$[key] = this.$[method](...args);
            } else if (item.captured) {
                const { key: key1 , captured  } = item;
                this.$[key1] = applyFluentCapture(captured, this.$);
            }
        }
    }
    copyMethods() {
        Object.assign(this.$, pick(this, [
            'atomic',
            'callback',
            'cleanup',
            'effect',
            'fulfill',
            'function',
            'mutate',
            'mutating',
            'reduce',
            'register',
            'render',
            'use',
            'using',
            'query',
            'when',
            'with',
            'withLock',
            'globalLock',
        ]));
    }
    runEffects(key) {
        this.effects.get(key).forEach(this.run);
    }
    clearKey(key) {
        this.effects.get(key).forEach((f)=>{
            delete f.values[key];
        });
    }
    // executeReducersFor(targetKey: keyof T) {
    //   for (const fx of this.effects.values()) {
    //       if (f.target === targetKey) {
    //       }
    //   }
    maybeAddChangeListener;
    createMemory() {
        const pending = this.target._pendingProperties;
        this.entries = entries(this.target);
        this.mem = fromEntries(this.entries.map(([key, value])=>{
            if (typeof value === 'symbol' && pending) {
                if (pending.has(value)) {
                    this.scheduled.add({
                        key,
                        value: pending.get(value)
                    });
                    value = void 0;
                }
            }
            if (typeof value === 'function' && value[FluentCaptureSymbol] === true) {
                this.scheduled.add({
                    key,
                    captured: value._results
                });
                value = void 0;
            }
            return [
                key,
                value
            ];
        }));
    }
    createAccessors() {
        //!? 'creating accessors'
        accessors(this.$, this.target, (key)=>({
                get: ()=>this.mem[key],
                set: (v)=>{
                    //!? 'context setter:', key, v
                    if (v === REDUCER) {
                        v = this.reducer.initial;
                        this.reducer.target = key;
                        // if value is filled first time then return and don't use initial
                        if (this.register(this.reducer)) return;
                        if (v == null) return;
                    }
                    this.target[key] = v;
                }
            }), (key)=>!this.keys.has(key));
        accessors(this.target, this.target, (key)=>({
                get: ()=>this.mem[key],
                set: (v)=>{
                    const prev = this.mem[key];
                    const isDifferent = !Object.is(prev, v);
                    //!? 'context target isDifferent', isDifferent, key, prev, v
                    // if (key === 'presets') console.log('is different', isDifferent, key, v)
                    if (isDifferent) {
                        this.mem[key] = v;
                        this.runEffects(key);
                    // this.maybeAddChangeListener(key, v)
                    }
                }
            }), (key)=>{
            if (!this.keys.has(key)) {
                this.keys.add(key);
                return true;
            }
            return false;
        });
        accessors(this.target, fromEntries(entries(this.mem).filter(([key, value])=>{
            if (this.maybeAddChangeListener(key, value)) return true;
        })), (key)=>({
                set: (v)=>{
                    //!? 'setting event target', key
                    this.maybeAddChangeListener(key, v);
                }
            }));
    }
    cleanup;
    register;
    update(f) {
        let changed = f.pass == null;
        f.pass ??= !f.keys.size;
        for (const key of f.keys){
            const value = this.target[key];
            if (value == null) {
                f.pass = false;
                return true;
            }
            if (!Object.is(f.values[key], value)) {
                f.values[key] = value;
                f.pass = true;
                changed = true;
            }
        }
        return changed;
    }
    flush;
    run;
    get _fulfill() {
        const f = this.active;
        const key = f?.target;
        if (f == null || key == null) {
            throw new TypeError('Attempted to use `fulfill` when no reducer was active.');
        }
        return (value)=>{
            this.target[key] = value;
        };
    }
    finalize() {
        const f = this.active;
        let result = f.fn.call(this, f.values);
        if (f.options?.isFulfillReducer) {
            result = result(this._fulfill);
        }
        if (result?.then) {
            result.then((res)=>{
                try {
                    if (f.cb?.(res) === false) return;
                    if (f.target != null && !f.options?.isFulfillReducer) {
                        this.target[f.target] = res;
                    }
                } finally{
                    if (f.options?.lock) {
                        this.active = null;
                        this.flush();
                    }
                }
            // if (typeof res === 'function') {
            //     res()
            //     f.dispose = null
            // }
            });
            if (!f.options?.lock) {
                this.active = null;
                this.flush();
            }
            return true;
        } else {
            return this.complete(result);
        }
    }
    complete;
    reduce;
    fulfill;
    effect;
    mutate;
    mutating;
    use;
    using;
    function;
    with;
    callback;
    atomic;
    query;
    when;
    withGlobalLock;
    withLock;
    target;
}

const components = new WeakMap();
if (!customElements.getName) {
    const tags = new WeakMap();
    const { define  } = customElements;
    customElements.define = function(name, ctor, options) {
        tags.set(ctor, name);
        return define.call(customElements, name, ctor, options);
    };
    customElements.getName = function(ctor) {
        return tags.get(ctor);
    };
}
let suffixKey = 0;
const fromElement = (ctor, options = ctor.elementOptions ?? {})=>{
    let fn = components.get(ctor);
    if (fn) return fn;
    let tag = customElements.getName(ctor);
    if (!tag) {
        tag = kebab(ctor.name).replace('-element', '');
        if (!tag.includes('-')) tag = 'x-' + tag;
        while(customElements.get(tag) != null){
            tag += (++suffixKey).toString(36);
        }
        customElements.define(tag, ctor, options);
    }
    if (options.extends) fn = (props)=>jsx(options.extends, Object.assign(props, {
            is: tag
        }), void 0);
    else fn = (props)=>jsx(tag, props, void 0);
    components.set(ctor, fn);
    fn.toString = ()=>tag;
    return fn;
};

const AttrTypes = new Map([
    [
        String,
        (x)=>x.toString()
    ],
    [
        Number,
        (x)=>parseFloat(x)
    ],
    [
        Boolean,
        (x)=>x = x === false ? false : x != null
    ],
]);
const applyAttrs = (self, data, initialUpdates)=>{
    for (const [k, v] of Object.entries(data)){
        //!? 'setting attr "%s"', k, v.toString()
        if ([
            String,
            Number,
            Boolean
        ].includes(v)) self[k] = data[k] = void 0;
        else if (v === true) data[k] = true;
        else if (typeof v === 'object') data[k] = v.toString();
    }
    const update = (attr, key, value)=>{
        if (typeof value === 'boolean') {
            if (value) {
                self.setAttribute(attr, '');
            } else {
                if (self.hasAttribute(attr)) {
                    self.removeAttribute(attr);
                } else {
                    data[key] = false;
                }
            }
        } else {
            //!? 'setting', attr, value
            self.setAttribute(attr, value);
        }
    };
    return accessors(self, data, (key)=>{
        const attr = key.toLowerCase();
        return {
            get: ()=>data[key],
            set (value) {
                //!warn 'setter changed', key, value
                if (!self.isConnected) initialUpdates.push(()=>update(attr, key, value));
                else update(attr, key, value);
            }
        };
    });
};
const attrListener = (body)=>new Function('event', `with(this){let fn=${body};return typeof fn=='function'?fn.call(this,event):fn}`);

class PropertySettings {
    attr = false;
    out = false;
    is;
}
const protoPropertyMap = new WeakMap();
const prop = toFluent(PropertySettings, (settings)=>(is)=>(proto, key)=>{
            if (typeof key !== 'string') return;
            settings.is ??= is;
            if (!protoPropertyMap.has(proto)) {
                protoPropertyMap.set(proto, new Map([
                    [
                        key,
                        settings
                    ]
                ]));
            } else {
                const propertyMap = protoPropertyMap.get(proto);
                if (propertyMap.has(key)) {
                    Object.assign(propertyMap.get(key), settings);
                } else {
                    propertyMap.set(key, settings);
                }
            }
        });
const attr = prop.attr;
const out = prop.out;
const is = prop.is;

const RefProxy = (target)=>Getter((key)=>({
            get current () {
                return target[key];
            },
            set current (el){
                target[key] = el;
            }
        }));
class Ref extends EventTarget {
    #current;
    key;
    constructor(ref){
        super();
        defineProperty.enumerable.get(()=>this.#current).set((el1)=>{
            this.#current = el1;
            dispatch(this, 'change');
        })(this, 'current');
        Object.assign(this, ref);
    }
}
class PropsRef {
    ref = new Ref();
    constructor(props, key){
        Object.assign(this, props);
        // TODO: implement ref.key in html-vdom
        this.ref.key = key;
    }
}
class RefCollection extends EventTarget {
    items = [];
    listeners = new Map();
    #construct = (param)=>{
        const item = this.construct(param);
        this.listeners.set(item, on(item.ref).change(()=>dispatch(this, 'change')));
        return item;
    };
    [Symbol.iterator]() {
        return this.items[Symbol.iterator]();
    }
    get length() {
        return this.items.length;
    }
    get refs() {
        return filterMap(this.items, (x)=>x.ref.current);
    }
    createItems(entriesOrValues = []) {
        this.items = entriesOrValues.map(this.#construct);
    }
    find(cb) {
        return this.refs.find(cb);
    }
    map(cb) {
        return this.items.map(cb);
    }
    push(state) {
        const result = this.items.push(this.#construct(state));
        dispatch(this, 'change');
        return result;
    }
    insertAfter(state, other) {
        const index = this.refs.indexOf(other);
        if (!~index) {
            throw new ReferenceError('Item not in ref collection.');
        }
        const newItem = this.#construct(state);
        this.splice(index + 1, 0, newItem);
        return newItem;
    }
    splice(start, deleteCount, ...args) {
        const result = this.items.splice(start, deleteCount, ...args);
        dispatch(this, 'change');
        return result;
    }
    move(oldIndex, newIndex) {
        const item = this.items.splice(oldIndex, 1)[0];
        this.splice(newIndex, 0, item);
    }
    add(...items) {
        for (const item of items){
            this.items.push(item);
        }
        dispatch(this, 'change');
    }
    getRefItem(item) {
        const index = this.refs.indexOf(item);
        if (!~index) {
            throw new ReferenceError('Item not in ref collection.');
        }
        return this.items[index];
    }
    remove(item) {
        const index = (item instanceof PropsRef ? this.items : this.refs).indexOf(item);
        if (!~index) {
            throw new ReferenceError('Item not in ref collection.');
        }
        return this.splice(index, 1)[0];
    }
}
class RefMap extends RefCollection {
    #map = new Map();
    construct([key, state]) {
        const item = new PropsRef(state, key);
        this.#map.set(item.ref.key, item);
        return item;
    }
    constructor(entries){
        super();
        this.createItems(entries);
    }
    toJSON() {
        return filterMap(this.items, (x)=>[
                x.ref.key,
                x.ref.current?.toJSON()
            ]);
    }
    get(key) {
        return this.#map.get(key);
    }
}
class RefSet extends RefCollection {
    construct(state) {
        return new PropsRef(state);
    }
    constructor(values){
        super();
        this.createItems(values);
    }
    toJSON() {
        return filterMap(this.items, (x)=>x.ref.current?.toJSON());
    }
}

function element(ctorOrOptions, options) {
    if (typeof ctorOrOptions === 'function') return fromElement(ctorOrOptions, options);
    if (typeof ctorOrOptions === 'object') options = ctorOrOptions;
    return (superclass)=>{
        const parent = superclass;
        // get own prototype decorated properties map
        const ownPropertyMap = protoPropertyMap.get(superclass.prototype) ?? new Map();
        const ownAttrKeys = [
            ...ownPropertyMap
        ].filter(([, settings])=>settings.attr).map(([key])=>key);
        // map from the lowercase (attribute) version to the actual property name
        const attrKeysMap = new Map(ownAttrKeys.map((x)=>[
                x.toLowerCase(),
                x
            ]));
        const attrTypes = new Map();
        // own observed attributes
        const ownObservedAttributes = [
            ...attrKeysMap.keys()
        ];
        //!? 'own observed attributes', ownObservedAttributes, superclass
        const ctor = class extends parent {
            static get elementOptions() {
                return options ?? super.elementOptions;
            }
            static get observedAttributes() {
                return ownObservedAttributes.concat(super.observedAttributes ?? []);
            }
            static get attributeKeys() {
                return ownAttrKeys.concat(super.attributeKeys ?? []);
            }
            static get propertyMap() {
                return new Map([
                    ...ownPropertyMap,
                    ...super.propertyMap ?? []
                ]);
            }
            static get exposedPropertyKeys() {
                return [
                    ...ctor.propertyMap
                ].filter(([, settings])=>settings.out).map(([key])=>key);
            }
            #isMain = true;
            #attrData;
            #attrInitialUpdates = [];
            $;
            host = this;
            isMounted = false;
            constructor(...args){
                super(...args);
                this.#isMain = new.target === ctor;
                if (this.#isMain) {
                    defineProperty.not.enumerable(this, '_ctor', ctor);
                    this.#create();
                    Object.assign(this, args[0]);
                    queueMicrotask(()=>{
                        this.#flushAttributeUpdates();
                    });
                }
            }
            toJSON() {
                return pick(this, ctor.exposedPropertyKeys);
            }
            // goto://context.ts#ContextClass.createMemory
            _scheduleProperty(key, ...args) {
                const symbol = Symbol();
                this._pendingProperties ??= new Map();
                this._pendingProperties.set(symbol, [
                    key,
                    args
                ]);
                return symbol;
            }
            _getPropertySettings(key) {
                return ctor.propertyMap.get(key);
            }
             #applyAttributes() {
                const attrs = pick(this, ctor.attributeKeys);
                for (const [key, value] of entries(attrs)){
                    const valueCtor = value?.constructor;
                    // TODO: maybe turn all as a .is constructor? Boolean needs special case
                    if (valueCtor?.from) {
                        const settings = ctor.propertyMap.get(key);
                        settings.is = valueCtor;
                    }
                    //!? 'value', value
                    const type = value == null ? AttrTypes.get(String) : AttrTypes.get(value) ?? (valueCtor ? AttrTypes.get(valueCtor) ?? (Object.hasOwn(valueCtor.prototype, 'toString') ? AttrTypes.get(String) : null) : null);
                    if (!type) {
                        throw new TypeError(`Attribute "${key}" is not valid type, must be either: String, Number, Boolean, null, undefined`);
                    }
                    //!? 'type', key, type.toString()
                    attrTypes.set(key, type);
                    attrKeysMap.set(key.toLowerCase(), key);
                }
                this.#attrData = attrs;
                applyAttrs(this, attrs, this.#attrInitialUpdates);
            }
             #flushAttributeUpdates() {
                this.#attrInitialUpdates.splice(0).forEach((fn)=>fn());
            }
             #attachReactiveContext() {
                ContextClass.attach(this, omit($, [
                    'transition'
                ]));
                this.host.on = $.on.bind(this, this);
                this.host.dispatch = $.dispatchBind(this);
            }
             #registerChangeEvents() {
                const keys = new Set(ctor.exposedPropertyKeys);
                this.$.effect.keys(keys).task(()=>{
                    $.dispatch.composed(this.host, 'change');
                });
            }
             #create() {
                //!? 'creating', this
                // attach reactive context
                this.#attachReactiveContext();
                this.#applyAttributes();
                // listen for lifecycle events
                $.on(this.host).mounted(()=>{
                    // we flush here as well as at the queue level
                    // because it can be created and mounted before
                    // the next tick, so it will be allowed to create
                    this.#flushAttributeUpdates();
                    // register 'change' events from @$.out() and @$.attr.out()
                    this.#registerChangeEvents();
                });
                $.on(this.host).unmounted(()=>{
                    this.$.cleanup();
                });
                // jsx render helpers
                this.$.render = this.$.effect.cb((result)=>{
                    render(result, this.root ?? (this.root = $.shadow(this)));
                    //!? 'render', this, result
                    return false;
                });
                this.$.part = toFluent(EffectOptions, (options)=>(fn, output)=>{
                        let update;
                        const cb = (value)=>{
                            output = value;
                            if (update) queueMicrotask(update);
                            return false;
                        };
                        const Fn = ()=>{
                            // lazily create effect when first used
                            if (!update) this.$.register(new Fx({
                                fn,
                                cb,
                                options,
                                dispose: cb
                            }));
                            update = hook;
                            return output;
                        };
                        return Fn;
                    });
                this.$.ref = RefProxy(this.$);
                this.created?.(this.$);
            }
            // based on: https://stackoverflow.com/a/49773201/175416
            dispatchEvent(event) {
                const onEventType = `on${event.type}`;
                let pass = true;
                if (this.#isMain && Object.hasOwn(this, onEventType)) {
                    let fn = this[onEventType];
                    if (fn) {
                        if (!fn) fn = attrListener(this.getAttribute(onEventType));
                        pass = fn.call(this, event);
                    }
                }
                // run prototype on<event>(event) listeners
                if (pass !== false) {
                    const fn1 = getOwnProperty(super.constructor.prototype, onEventType);
                    if (fn1) pass = fn1.call(this, event);
                }
                // run prototype <event>(context, event) listeners
                if (pass !== false) {
                    const fn2 = getOwnProperty(super.constructor.prototype, event.type);
                    if (fn2) pass = fn2.call(this, this.$, event);
                }
                if (pass !== false) super.dispatchEvent(event);
                return pass;
            }
            connectedCallback() {
                if (!this.#isMain) {
                    super.connectedCallback?.();
                    return;
                }
                //!? 'connected'
                super.connectedCallback?.();
                if (!this.isMounted) {
                    this.isMounted = true;
                    this.host.dispatch('mounted');
                }
            }
            disconnectedCallback() {
                if (!this.#isMain) {
                    super.disconnectedCallback?.();
                    return;
                }
                super.disconnectedCallback?.();
                // When the element is moved, rather than removed,
                // By observing the state at the end of the task
                // we only fire unmounted when it's been actually
                // immediately.
                queueMicrotask(()=>{
                    if (!this.isConnected) {
                        this.isMounted = false;
                        this.host.dispatch('unmounted');
                    }
                });
            }
            attributeChangedCallback(name, oldValue, newValue) {
                if (!this.#isMain) {
                    super.attributeChangedCallback?.(name, oldValue, newValue);
                    return;
                }
                //!warn 'attributeChangedCallback:', name, newValue
                let key;
                if (key = attrKeysMap.get(name)) {
                    const { is  } = ctor.propertyMap.get(key);
                    const prev = this.#attrData[key];
                    const next = attrTypes.get(key)(newValue);
                    // type casting comparison for objects that serialize their attribute value
                    const isDifferent = is ? prev != next : !Object.is(prev, next);
                    if (isDifferent) {
                        this.#attrData[key] = next;
                        this[key] = is ? this[key] == next ? this[key] : is.from(next) : next;
                        //!? 'propertyChangedCallback:', key, next
                        this.propertyChangedCallback?.(key, prev, next);
                    }
                }
                super.attributeChangedCallback?.(name, oldValue, newValue);
            }
        };
        // name the anonymous class with the decorated class' original name
        defineProperty(ctor, 'name', superclass.name);
        return ctor;
    };
}

class SlottedSettings {
    elements = bool;
    nodes = bool;
    firstChild = bool;
    deep = bool;
}
const slotted$1 = (c)=>toFluent(SlottedSettings, (settings)=>(mapFn = (x)=>x)=>c._scheduleProperty('fulfill', ({ root  })=>(cb)=>$.onSlotChange(root, (slotted)=>{
                        let result;
                        if (settings.nodes) result = slotted.nodes;
                        else if (settings.firstChild) result = [
                            slotted.firstChild
                        ];
                        else result = slotted.elements;
                        if (settings.deep) {
                            result = result.flatMap((x)=>x instanceof HTMLSlotElement ? [
                                    ...x.assignedElements()
                                ] : x);
                        }
                        result = filterMap(result, mapFn);
                        if (!result.length) cb(void 0);
                        else if (settings.firstChild) cb(result[0]);
                        else cb(result);
                    })));

class State extends EventTarget {
    static from(x) {
        return new State(Array.isArray(x) ? x : x.toString().split(','));
    }
    stack;
    constructor(states){
        super();
        this.states = states;
        this.stack = [
            states.Idle
        ];
        // @ts-ignore
        $.on(this).change(()=>{
            $.dispatch(this, 'statechange', this);
        });
        if (this.stack.length) {
            $.dispatch(this, 'change');
        }
    }
    toJSON() {
        return this.states;
    }
    get current() {
        return this.toString();
    }
    get isIdle() {
        return this.stack.length === 1;
    }
    toString() {
        return this.stack.at(-1);
    }
    is(testState) {
        return this.stack.at(-1) === testState;
    }
    emit(state, ...detail) {
        if (!this.is(state)) {
            throw new TypeError(`Attempt to emit for "${state}" but our current state stack is "${this.stack}"`);
        }
        $.dispatch(this, state, ...detail);
    }
    push(state, ...detail) {
        if (this.is(state)) {
            throw new TypeError(`Attempt to push state "${state}" again on top of itself, our current state stack is "${this.stack}"`);
        }
        this.stack = [
            ...this.stack,
            state
        ];
        //!warn 'push %s', this.stack
        $.dispatch(this, `${this.stack.at(-2)}pause`);
        $.dispatch(this, `${state}start`, ...detail);
        $.dispatch(this, 'change');
    }
    pop(state, ...detail) {
        if (!this.is(state)) {
            throw new TypeError(`Attempt to pop state "${state}" but our current state stack is "${this.stack}"`);
        }
        if (this.stack.length === 1) {
            throw new TypeError(`Attempt to pop base state "${state}" when stack is length 1.`);
        }
        this.stack = [
            ...this.stack.slice(0, -1)
        ];
        //!warn 'pop %s (-%s)', this.stack, state
        $.dispatch(this, `${state}end`, ...detail);
        $.dispatch(this, `${this.stack.at(-1)}resume`);
        $.dispatch(this, 'change');
    }
    cancel(state, ...detail) {
        if (!this.is(state)) {
            throw new TypeError(`Attempt to cancel state "${state}" but our current state stack is "${this.stack}"`);
        }
        if (this.stack.length === 1) {
            throw new TypeError(`Attempt to cancel base state "${state}" when stack is length 1.`);
        }
        this.stack = [
            ...this.stack.slice(0, -1)
        ];
        //!warn 'cancel %s (-%s)', this.stack, state
        $.dispatch(this, `${state}cancel`, ...detail);
        $.dispatch(this, `${this.stack.at(-1)}resume`);
        $.dispatch(this, 'change');
    }
    swap(state, ...detail) {
        if (this.stack.length === 1) {
            throw new TypeError(`Attempt to swap base state "${state}" when stack is length 1.`);
        }
        const prev = this.current;
        this.stack = [
            ...this.stack.slice(0, -1),
            state
        ];
        $.dispatch(this, `${prev}cancel`);
        $.dispatch(this, `${this.stack.at(-1)}start`, ...detail);
        $.dispatch(this, 'change');
    }
    returnToIdle() {
        while(!this.isIdle){
            this.cancel(this.current, void 0);
        }
    }
    states;
}

/**
 * Token interface
 */ class Token extends String {
    static create(value, group, source) {
        return new Token(value, group, source);
    }
    /**
   * The group it matched.
   */ group;
    /**
   * The value of the match.
   */ get value() {
        return '' + this;
    }
    /**
   * The index position of the match.
   */ get index() {
        return this.source.index;
    }
    /**
   * The input string.
   */ source;
    constructor(value, group, source){
        super(value);
        this.group = group;
        Object.defineProperty(this, 'source', {
            enumerable: false,
            value: source
        });
    }
    is(group, value) {
        return group == this.group && (value == null || value == this.value);
    }
    as(value, group = this.group) {
        return Token.create(value, group, this.source);
    }
}

/**
 * A finder function that returns first entry that is not nullish.
 *
 * @param e An entry tuple as returned by Object.entries()
 *
 * @private
 */ const NonNull = (e)=>e[1] != null;

/**
 * Convert a match object from `string.match()` or `string.matchAll()`
 * with a named group RegExp such as `/(?<group>[a-z])/` to a more
 * useful {@link Token} object.
 *
 * @param match The match object
 */ const matchToToken = (match)=>{
    if (!match) return null;
    if (match.index == null) return null;
    if (!match.groups) throw new TypeError('RegExp match is missing named groups such as: /(?<group>[a-z])/');
    const entry = Object.entries(match.groups).find(NonNull);
    if (entry) return Token.create(entry[1], entry[0], match);
    return null;
};

/**
 * Utility to join RegExps together.
 * @private
 */ const joinRegExps = (regexps)=>new RegExp(regexps.map((x)=>x.toString()).map((x)=>x.slice(x.indexOf('/') + 1, x.lastIndexOf('/'))).join('|'), 'gi');

// eslint-disable-next-line @typescript-eslint/no-unused-vars
/**
 * Create a {@link TokenizerFactory} for the given RegExps.
 *
 * To capture, RegExps must use a [named group](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions/Groups_and_Ranges#using_named_groups).
 *
 * ```ts
 * const tokenize = createTokenizer(
 *   /(?<ident>[a-z]+)/, // named groups determine token `group`
 *   /(?<number>[0-9]+)/
 * )
 * ```
 *
 * @param regexps RegExps to match.
 */ const createTokenizer = (...regexps)=>{
    const regexp = joinRegExps(regexps);
    return (input)=>{
        const matches = input.matchAll(regexp);
        const next = ()=>matchToToken(matches.next().value);
        const iterator = function*(token) {
            while(token = next())yield token;
        };
        Object.defineProperty(next, Symbol.iterator, {
            value: iterator
        });
        return next;
    };
};

const extraWhitespace = /\s+/g;
const tokenizer = createTokenizer(/(?<open>\s*\{\s*)/, /(?<close>\s*\}\s*)/, /(?<comment>\/\*[^]*?\*\/)/, /\s*(?<rule>[^{;}/'"]+)\s+?(?={)/, /\s*(?<string>'.*?'|".*?")(?=;)/, /\s*(?<prop>[^:;/]+)(?=:)/, /\s*(?<value>[^:;]+)(?=;)/);
/**
 * Convert a CSS string to a {@link NestedCSSDeclaration}.
 *
 * @param input
 */ function cssToJs(input) {
    input = input.replace(extraWhitespace, ' ');
    const nextToken = tokenizer(input);
    const filter = (ignored)=>()=>{
            let token;
            while(token = nextToken())if (!ignored.includes(token.group)) break;
            return token;
        };
    const next = filter([
        'comment'
    ]);
    const parse = (style = {})=>{
        let token;
        while(token = next()){
            const { value , group  } = token;
            switch(group){
                case 'rule':
                    next() // open {
                    ;
                    // @ts-ignore can't handle the ??= for some reason
                    parse(style[value.trim()] ??= {});
                    break;
                case 'prop':
                    style[value.trim()] = next()?.value?.trim();
                    break;
                case 'close':
                    return style;
            }
        }
        return style;
    };
    return parse();
}

/**
 * Joins parts with values
 *
 * @param parts
 * @param values
 */ function joinPartsWithValues(parts, values) {
    let str = '';
    for(let i = 0; i < parts.length; i++){
        str += parts[i];
        // if (i < values.length) str += values[i]
        str += values[i];
    }
    return str;
}
/**
 * Convert string to kebab-case.
 *
 * @param s PascalCase or camelCase string
 */ function kebabCase(s) {
    return s.replace(/[A-Z]/g, (m, i)=>(i ? '-' : '') + m).toLowerCase();
}

// original: https://gist.github.com/mindplay-dk/47bc597dcbc28bea009b02c5a0c9010e
const mediaRule = (rule, media)=>media ? `${media}{${rule}}` : rule;
const createRule = (target, prop, value)=>`${target}{${kebabCase(prop)}:${value}}`;
const remap = (target, map)=>// isClassOrRoot.test(target)
    //   :
    [
        ...target.matchAll(/[\w-]+|./g)
    ].flat().map((x)=>map?.get(x) ?? x).join('');
/**
 * Compile a JS nested rules {@link NestedCSSDeclaration} to a CSS string.
 *
 * Examples:
 * ```ts
 * jsToCss({ '.foo': { color: 'blue' } })
 * // => .foo{color:blue}
 *
 * // custom root
 * jsToCss({ color: 'red' }, '.my-button')
 * // => .my-button{color:red}
 *
 * // with substitution
 * jsToCss({ '.foo': { color: 'blue' } }, null, new Map([['foo', 'bar']]))
 * // => .bar{color:blue}
 * ```
 *
 * @param rules Rules object
 * @param rootSelector Top level rules will use this selector
 * @param aliasMap Alias identifiers (i.e for `.foo` to become `.bar`
 *  you will need a `foo=bar` entry)
 * @returns The compiled CSS string
 */ function jsToCss(rules, rootSelector, aliasMap) {
    rootSelector ??= '';
    let css = '';
    const parse = (rules, child = '', media = '')=>{
        for(const key in rules){
            const value = rules[key];
            if (typeof value === 'object') {
                const isMedia = /^@/.test(key) ? key : null;
                const isPct = /%$/.test(key);
                let target = child;
                if (isPct) target = key;
                else if (!isMedia) {
                    target = key.split(',').map((k)=>k.trim()).map((k)=>child.split(',').map((x)=>(x + (/^&/.test(k) ? k.slice(1) : ' ' + k)).trim()).join(',')).map((k)=>k.trim()).join(',');
                }
                parse(value, target, isMedia || media);
            } else {
                const rule = mediaRule(createRule(child ? remap(child, aliasMap) : rootSelector, key, value), media);
                css += '\n' + rule.trim();
            }
        }
    };
    parse({
        [rootSelector]: rules
    });
    return css.trim();
}

/**
 * Factory a {@link NestedCSSCompiler} for the given string.
 *
 * @param parts
 * @param values
 */ function css(parts, ...values) {
    const nestedCssString = joinPartsWithValues(parts, values);
    const nestedCssObject = cssToJs(nestedCssString);
    function compileCss(rootSelector = ':host', aliasMap) {
        return jsToCss(nestedCssObject, rootSelector, aliasMap);
    }
    compileCss.valueOf = compileCss;
    return compileCss;
}

/******************************************************************************
Copyright (c) Microsoft Corporation.
Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.
THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}
function __spreadArray(to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
}

/**
 * Helper function to apply arguments to a selected function with type-inference for the supplied function.
 * For instance, in the following statement `applyArg(1).to((x) => x + 1)`, the type of `x` is correctly inferred to be `number`
 * @param args the argument list to be supplied... the argument list of the function supplied to the `to` clause (see below) should match this list
 * @returns an object with a single method, `.to(f)` where `f` is the function to which param `(...args)` will be passed.
 * The return type of this method will match the return type of `f`
 */
var applyArgs = function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var f = (function (f) { return f.apply(void 0, __spreadArray([], __read(args), false)); });
    f.to = f;
    return f;
};
function pipeImpl() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var _a = __read(args), o1 = _a[0], operations = _a.slice(1);
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return operations.reduce(function (acc, f) { return f(acc); }, o1.apply(void 0, __spreadArray([], __read(args), false)));
    };
}
function pipeInto(src, o1) {
    var operations = [];
    for (var _i = 2; _i < arguments.length; _i++) {
        operations[_i - 2] = arguments[_i];
    }
    return applyArgs(src).to(pipeImpl.apply(void 0, __spreadArray([o1], __read(operations), false)));
}

var Intersect;
(function(Intersect) {
    Intersect[Intersect["None"] = 0] = "None";
    Intersect[Intersect["Left"] = 1] = "Left";
    Intersect[Intersect["Top"] = 2] = "Top";
    Intersect[Intersect["Right"] = 4] = "Right";
    Intersect[Intersect["Bottom"] = 8] = "Bottom";
    Intersect[Intersect["Inside"] = 16] = "Inside";
})(Intersect || (Intersect = {}));

class Scalar {
    static interpolate(a, b, t) {
        return a + (b - a) * t;
    }
    static radiansToDegrees(radians) {
        return radians * 180 / Math.PI;
    }
    static degreesToRadians(degrees) {
        return degrees * Math.PI / 180;
    }
}

class Shape {
    x;
    y;
    get left() {
        return this.x;
    }
    set left(x) {
        this.x = x;
    }
    get top() {
        return this.y;
    }
    set top(y) {
        this.y = y;
    }
    get bottom() {
        return this.y;
    }
    set bottom(y) {
        this.y = y;
    }
    get right() {
        return this.x;
    }
    set right(x) {
        this.x = x;
    }
    clone() {
        const ctor = this.constructor;
        return new ctor(this);
    }
    toJSON() {
        return this;
    }
    draw(color = 'red', position = 'absolute') {
        const div = document.createElement('div');
        Object.assign(div.style, {
            position,
            boxSizing: 'border-box',
            border: '1px solid ' + color,
            ...this.toStyle(),
            zIndex: 1000
        });
        document.body.appendChild(div);
        return div;
    }
    screen(other = this) {
        return this.clone().screenSelf(other);
    }
    screenSelf(other = this) {
        return this.translateSelf(other.negate());
    }
    negate() {
        return this.clone().negateSelf();
    }
    negateSelf() {
        this.x = -this.x;
        this.y = -this.y;
        return this;
    }
    contain(other) {
        return this.clone().containSelf(other);
    }
    containSelf(other) {
        if (this.top < other.top) this.top = other.top;
        else if (this.bottom > other.bottom) this.bottom = other.bottom;
        if (this.right > other.right) this.right = other.right;
        else if (this.left < other.left) this.left = other.left;
        return this;
    }
    scale(x = 0, y = x) {
        return this.clone().scaleSelf(x, y);
    }
    scaleSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            this.width *= x.width;
            this.height *= x.height;
        } else {
            this.width *= x;
            this.height *= y;
        }
        return this;
    }
    scaleLinear(x = 0, y = x) {
        return this.clone().scaleLinearSelf(x, y);
    }
    scaleLinearSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            this.width += x.width;
            this.height += x.height;
        } else {
            this.width += x;
            this.height += y;
        }
        return this;
    }
    zoomLinear(x = 0, y = x) {
        return this.clone().zoomLinearSelf(x, y);
    }
    zoomLinearSelf(x = 0, y = x) {
        if (x instanceof Shape) {
            this.scaleLinearSelf(x).translateSelf(x.scale(-0.5));
        } else {
            this.scaleLinearSelf(x, y).translateSelf(x * -0.5, y * -0.5);
        }
        return this;
    }
    add(x = 0, y = x) {
        return this.translate(x, y);
    }
    addSelf(x = 0, y = x) {
        return this.translateSelf(x, y);
    }
    sub(x = 0, y = x) {
        return this.clone().subSelf(x, y);
    }
    subSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            return this.translateSelf(x.negate());
        } else {
            return this.translateSelf(-x, -y);
        }
    }
    translate(x = 0, y = x) {
        return this.clone().translateSelf(x, y);
    }
    translateSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            this.x += x.x || 0;
            this.y += x.y || 0;
        } else {
            this.x += x;
            this.y += y;
        }
        return this;
    }
    touchPoint(other, center = this.center ?? this) {
        const self = this instanceof Rect ? this : new Point(1, 1);
        const i = this.intersectPoint(other, center).translateSelf(other.center);
        const x = i.x - self.width * 0.5;
        const y = i.y - self.height * 0.5;
        return new Point(x, y);
    }
    intersectPoint(other, center = this.center ?? this) {
        const self = this instanceof Rect ? this : new Point(1, 1);
        const w = (self.width + other.width) * 0.5;
        const h = (self.height + other.height) * 0.5;
        const d = center.screen(other.center);
        const tan_phi = h / w;
        const tan_theta = Math.abs(d.y / d.x);
        // tell me in which quadrant the A point is
        const qx = Math.sign(d.x);
        const qy = Math.sign(d.y);
        let xI, yI;
        if (tan_theta > tan_phi) {
            xI = h / tan_theta * qx;
            yI = h * qy;
        } else {
            xI = w * qx;
            yI = w * tan_theta * qy;
        }
        return new Point(xI, yI);
    }
    toStylePosition() {
        return {
            left: this.x + 'px',
            top: this.y + 'px'
        };
    }
    toStyleSize() {
        return {
            width: this.width + 'px',
            height: this.height + 'px'
        };
    }
    toStyle() {
        return {
            ...this.toStylePosition(),
            ...this.toStyleSize()
        };
    }
    toPositionObject() {
        return {
            x: this.x,
            y: this.y
        };
    }
    toSizeObject() {
        return {
            width: this.width,
            height: this.height
        };
    }
}

class Point extends Shape {
    static fromElement(el) {
        return new Point(+(el.dataset.left ?? el.offsetLeft), +(el.dataset.top ?? el.offsetTop));
    }
    static fromObject(obj) {
        return new Point(obj.x ?? obj.width, obj.y ?? obj.height);
    }
    static fromMatrix(matrix) {
        return new Point(matrix.e, matrix.f);
    }
    static fromAngle(radians) {
        return new Point(Math.cos(radians), Math.sin(radians));
    }
    static fromAngleDegrees(degrees) {
        return Point.fromAngle(Scalar.degreesToRadians(degrees));
    }
    x;
    y;
    constructor(x = 0, y = x){
        super();
        if (typeof x === 'object') Object.assign(this, Point.fromObject(x.toJSON ? x.toJSON() : x));
        else {
            this.x = x;
            this.y = y;
        }
    }
    [Symbol.iterator]() {
        return [
            this.x,
            this.y
        ][Symbol.iterator]();
    }
    toString() {
        return `${this.x} ${this.y}`;
    }
    set(other) {
        this.x = other.x;
        this.y = other.y;
        return this;
    }
    draw(color = 'yellow', position = 'absolute') {
        const div = document.createElement('div');
        Object.assign(div.style, {
            position,
            boxSizing: 'border-box',
            border: '1px solid ' + color,
            ...this.toStylePosition(),
            width: '2px',
            height: '2px',
            zIndex: 1000
        });
        document.body.appendChild(div);
        return div;
    }
    get pos() {
        return new Point(this.x, this.y);
    }
    get position() {
        return this.pos;
    }
    get size() {
        return new Point(this.width, this.height);
    }
    get width() {
        return this.x;
    }
    set width(x) {
        this.x = x;
    }
    get height() {
        return this.y;
    }
    set height(y) {
        this.y = y;
    }
    interpolate(other, t) {
        return this.clone().interpolateSelf(other, t);
    }
    interpolateSelf(other, t) {
        this.x = this.x + (other.x - this.x) * t;
        this.y = this.y + (other.y - this.y) * t;
        return this;
    }
    diff(other) {
        return this.clone().diffSelf(other);
    }
    diffSelf(other) {
        this.x -= other.x;
        this.y -= other.y;
        return this;
    }
    abs() {
        return this.clone().absSelf();
    }
    absSelf() {
        this.x = Math.abs(this.x);
        this.y = Math.abs(this.y);
        return this;
    }
    square() {
        return this.clone().squareSelf();
    }
    squareSelf() {
        this.x *= this.x;
        this.y *= this.y;
        return this;
    }
    sum() {
        return this.x + this.y;
    }
    manhattan(other) {
        return this.diff(other).absSelf().sum();
    }
    octile(other) {
        const d = this.diff(other).absSelf();
        const F = Math.SQRT2 - 1;
        return d.x < d.y ? F * d.x + d.y : F * d.y + d.x;
    }
    max() {
        return Math.max(this.x, this.y);
    }
    min() {
        return Math.min(this.x, this.y);
    }
    chebyshev(other) {
        return this.diff(other).absSelf().max();
    }
    euclidean(other) {
        return this.distance(other);
    }
    distance(other) {
        return this.diff(other).mag();
    }
    mag() {
        return Math.sqrt(this.square().sum());
    }
    length() {
        return this.mag();
    }
    unit() {
        return this.scale(1 / this.mag());
    }
    dot(other) {
        return new Line(this, other).dot();
    }
    normal() {
        return new Point(this.y, -this.x);
    }
    angleTo(other) {
        return new Line(this, other).angle();
    }
    round() {
        return this.clone().roundSelf();
    }
    roundSelf() {
        this.x = Math.round(this.x);
        this.y = Math.round(this.y);
        return this;
    }
    precisionRound(p = 1) {
        return this.clone().precisionRoundSelf(p);
    }
    precisionRoundSelf(p = 1) {
        this.x = Math.round(this.x * p) / p;
        this.y = Math.round(this.y * p) / p;
        return this;
    }
    gridRound(p = 1) {
        return this.clone().gridRoundSelf(p);
    }
    gridRoundSelf(p = 1) {
        this.x = Math.round(this.x / p) * p;
        this.y = Math.round(this.y / p) * p;
        return this;
    }
    absoluteSum() {
        return Math.abs(this.x) + Math.abs(this.y);
    }
    withinRect(other) {
        return this.x >= other.left && this.x <= other.right && this.y >= other.top && this.y <= other.bottom;
    }
    transform(matrix) {
        return this.clone().transformSelf(matrix);
    }
    transformSelf(matrix) {
        const { x , y  } = this;
        const { a , b , c , d , e , f  } = matrix;
        this.x = a * x + c * y + e;
        this.y = b * x + d * y + f;
        return this;
    }
    multiply(x = 0, y = x) {
        return this.clone().multiplySelf(x, y);
    }
    multiplySelf(x = 0, y = x) {
        if (x instanceof Shape) {
            this.x *= x.x;
            this.y *= x.y;
        } else {
            this.x *= x;
            this.y *= y;
        }
        return this;
    }
    normalize(x, y = x) {
        return this.clone().normalizeSelf(x, y);
    }
    normalizeSelf(x = this.mag(), y = x) {
        if (x instanceof DOMMatrix) {
            this.x = (this.x - x.e) / x.a;
            this.y = (this.y - x.f) / x.d;
        } else if (typeof x === 'number') {
            this.x /= x;
            this.y /= y;
        } else {
            this.x /= x.x;
            this.y /= x.y;
        }
        return this;
    }
    equals(other) {
        return this.x === other.x && this.y === other.y;
    }
    equalsAny(other) {
        return this.x === other.x || this.y === other.y;
    }
}

class Line {
    p1;
    p2;
    constructor(p1, p2){
        if (p1.p1) {
            this.p1 = new Point(p1.p1);
            this.p2 = new Point(p1.p2);
        } else {
            this.p1 = p1;
            this.p2 = p2;
        }
    }
    [Symbol.iterator]() {
        return this.points[Symbol.iterator]();
    }
    get points() {
        return [
            this.p1,
            this.p2
        ];
    }
    angle() {
        const dx = this.p2.x - this.p1.x;
        const dy = this.p2.y - this.p1.y;
        return Math.atan2(dy, dx);
    }
    angleDegrees() {
        return Scalar.radiansToDegrees(this.angle());
    }
    clone() {
        const ctor = this.constructor;
        return new ctor(this);
    }
    mag() {
        return this.p1.distance(this.p2);
    }
    dot() {
        return this.p1.x * this.p2.x + this.p1.y * this.p2.y;
    }
    intersectionRect(r) {
        const p1 = this.p1;
        const p2 = this.p2;
        const is = (this.intersectsLine(r.leftLine) ? Intersect.Left : 0) + (this.intersectsLine(r.topLine) ? Intersect.Top : 0) + (this.intersectsLine(r.rightLine) ? Intersect.Right : 0) + (this.intersectsLine(r.bottomLine) ? Intersect.Bottom : 0) + (p1.withinRect(r) && p2.withinRect(r) ? Intersect.Inside : 0);
        return is;
    }
    intersectsRect(r) {
        const p1 = this.p1;
        const p2 = this.p2;
        // dprint-ignore
        return this.intersectsLine(r.leftLine) ? Intersect.Left : this.intersectsLine(r.topLine) ? Intersect.Top : this.intersectsLine(r.rightLine) ? Intersect.Right : this.intersectsLine(r.bottomLine) ? Intersect.Bottom : p1.withinRect(r) && p2.withinRect(r) ? Intersect.Inside : Intersect.None;
    }
    intersectsLine(other) {
        const a1 = this.p1;
        const a2 = this.p2;
        const b1 = other.p1;
        const b2 = other.p2;
        let q = (a1.y - b1.y) * (b2.x - b1.x) - (a1.x - b1.x) * (b2.y - b1.y);
        const d = (a2.x - a1.x) * (b2.y - b1.y) - (a2.y - a1.y) * (b2.x - b1.x);
        if (d == 0) {
            return false;
        }
        const r = q / d;
        q = (a1.y - b1.y) * (a2.x - a1.x) - (a1.x - b1.x) * (a2.y - a1.y);
        const s = q / d;
        if (r < 0 || r > 1 || s < 0 || s > 1) {
            return false;
        }
        return true;
    }
    // collide line to rectangle and return a new line
    // that is placed just outside of the rectangle and not within
    getLineToRectangleCollisionResponse(intersection, r) {
        if (intersection === Intersect.None) {
            return this;
        }
        const p1 = this.p1;
        const p2 = this.p2;
        const a1 = p1.clone();
        const a2 = p2.clone();
        const b1 = r.topLeft;
        const b2 = r.bottomRight;
        if (intersection & Intersect.Left) {
            a1.x = b1.x - 1;
        }
        if (intersection & Intersect.Top) {
            a1.y = b1.y - 1;
        }
        if (intersection & Intersect.Right) {
            a2.x = b2.x + 1;
        }
        if (intersection & Intersect.Bottom) {
            a2.y = b2.y + 1;
        }
        if (intersection & Intersect.Inside) {
            const p = p1.interpolate(p2, 0.5);
            const tp = p.touchPoint(r);
            a1.x = tp.x;
            a1.y = tp.y;
            a2.x = tp.x;
            a2.y = tp.y;
        }
        //   a1.x = tp.x
        //   a1.y = tp.y
        //   a2.x = tp.x
        //   a2.y = tp.y
        return new Line(a1, a2);
    }
    translate(x = 0, y = x) {
        return this.clone().translateSelf(x, y);
    }
    translateSelf(x = 0, y = x) {
        this.p1.translateSelf(x, y);
        this.p2.translateSelf(x, y);
        return this;
    }
}

class Matrix extends DOMMatrix {
    constructor(matrix){
        if (typeof matrix === 'string') {
            super();
        } else if (Array.isArray(matrix)) {
            super(matrix);
        } else if (typeof matrix === 'object') {
            super(Object.values(matrix));
        } else {
            super(matrix);
        }
    }
    toJSON() {
        return this.toFloat64Array();
    }
}

class Polyline {
    static fromPoints(points) {
        return new Polyline(points.slice(0, -1).map((p, i)=>{
            const next = points[i + 1];
            const line = new Line(p, next);
            return line;
        }));
    }
    lines;
    constructor(lines){
        if (lines.lines) {
            this.lines = lines.lines;
        } else {
            this.lines = lines;
        }
    }
    get normals() {
        const mags = this.lines.map((line)=>line.mag());
        const total = mags.reduce((a, b)=>a + b, 0);
        const normals = mags.map((mag)=>mag / total);
        return normals;
    }
    get path() {
        const points = [];
        for (const line of this.lines){
            points.push(line.p1);
        }
        points.push(this.lines.at(-1).p2);
        return new Polygon(points);
    }
    get length() {
        return this.lines.reduce((a, b)=>a + b.mag(), 0);
    }
    chopAt(index) {
        const points = [];
        for (const [i, line] of this.lines.entries()){
            points.push(line.p1);
            if (i === index) {
                const np = line.p1.interpolate(line.p2, 0.5);
                points.push(np);
            }
        }
        points.push(this.lines.at(-1).p2);
        this.lines = Polyline.fromPoints(points).lines;
        return this;
    }
}

class Rect extends Shape {
    static fromElement(el) {
        return new Rect(+(el.dataset.x ?? el.offsetLeft), +(el.dataset.y ?? el.offsetTop), +(el.dataset.width ?? el.offsetWidth), +(el.dataset.height ?? el.offsetHeight));
    }
    static fromObject(obj) {
        return new Rect(obj.x ?? obj.width, obj.y ?? obj.height, obj.width ?? 0, obj.height ?? 0);
    }
    static fromPoints(topLeft, bottomRight) {
        return new Rect(topLeft.x, topLeft.y, bottomRight.x - topLeft.x, bottomRight.y - topLeft.y);
    }
    static fromUnsortedPoints(p1, p2) {
        const topLeft = new Point(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y));
        const bottomRight = new Point(Math.max(p1.x, p2.x), Math.max(p1.y, p2.y));
        return new Rect(topLeft.x, topLeft.y, bottomRight.x - topLeft.x, bottomRight.y - topLeft.y);
    }
    static combine(rects) {
        const x = Math.min(...rects.map((b)=>b.x));
        const y = Math.min(...rects.map((b)=>b.y));
        return new Rect(x, y, Math.max(...rects.map((b)=>b.right)) - x, Math.max(...rects.map((b)=>b.bottom)) - y);
    }
    x;
    y;
    width;
    height;
    #stale = true;
    constructor(x = 0, y = x, width = x, height = y){
        super();
        if (typeof x === 'object') Object.assign(this, Rect.fromObject(x.toJSON ? x.toJSON() : x));
        else {
            this.x = x;
            this.y = y;
            this.width = width;
            this.height = height;
        }
    }
    [Symbol.iterator]() {
        return [
            this.x,
            this.y,
            this.width,
            this.height
        ][Symbol.iterator]();
    }
    toJSON() {
        return {
            x: this.x,
            y: this.y,
            width: this.width,
            height: this.height
        };
    }
    toString() {
        return `${this.x} ${this.y} ${this.width} ${this.height}`;
    }
    draw(color = 'red', position = 'absolute') {
        const div = document.createElement('div');
        Object.assign(div.style, {
            position,
            boxSizing: 'border-box',
            border: '1px solid ' + color,
            ...this.toStyle(),
            zIndex: 1000
        });
        document.body.appendChild(div);
        return div;
    }
    set(other) {
        this.x = other.x;
        this.y = other.y;
        this.width = other.width;
        this.height = other.height;
        this.#stale = true;
        return this;
    }
    setPosition(other) {
        this.x = other.x;
        this.y = other.y;
        this.#stale = true;
        return this;
    }
    setSize(other) {
        this.width = other.width;
        this.height = other.height;
        this.#stale = true;
        return this;
    }
    get points() {
        return [
            new Point(this.x, this.y),
            new Point(this.right, this.y),
            new Point(this.right, this.bottom),
            new Point(this.x, this.bottom),
        ];
    }
    get pos() {
        return new Point(this.x, this.y);
    }
    get position() {
        return this.pos;
    }
    get size() {
        return new Point(this.width, this.height);
    }
    get center() {
        return new Point(this.x + this.width * 0.5, this.y + this.height * 0.5);
    }
    get topLeft() {
        return this.pos;
    }
    get topRight() {
        return new Point(this.right, this.top);
    }
    get bottomLeft() {
        return new Point(this.x, this.bottom);
    }
    get bottomRight() {
        return this.pos.translate(this.size);
    }
    get right() {
        return this.x + this.width;
    }
    set right(x) {
        this.x = x - this.width;
    }
    get bottom() {
        return this.y + this.height;
    }
    set bottom(y) {
        this.y = y - this.height;
    }
    #leftLine;
    #topLine;
    #rightLine;
    #bottomLine;
    maybeCalculateLines() {
        if (!this.#stale) return;
        this.#leftLine = new Line(this.topLeft, this.bottomLeft);
        this.#topLine = new Line(this.topLeft, this.topRight);
        this.#rightLine = new Line(this.topRight, this.bottomRight);
        this.#bottomLine = new Line(this.bottomLeft, this.bottomRight);
        this.#stale = false;
    }
    get leftLine() {
        this.maybeCalculateLines();
        return this.#leftLine;
    }
    get topLine() {
        this.maybeCalculateLines();
        return this.#topLine;
    }
    get rightLine() {
        this.maybeCalculateLines();
        return this.#rightLine;
    }
    get bottomLine() {
        this.maybeCalculateLines();
        return this.#bottomLine;
    }
    interpolate(other, t) {
        return this.clone().interpolateSelf(other, t);
    }
    interpolateSelf(other, t) {
        this.x = this.x + (other.x - this.x) * t;
        this.y = this.y + (other.y - this.y) * t;
        this.width = this.width + (other.width - this.width) * t;
        this.height = this.height + (other.height - this.height) * t;
        return this;
    }
    round() {
        return this.clone().roundSelf();
    }
    roundSelf() {
        this.x = Math.round(this.x);
        this.y = Math.round(this.y);
        this.width = Math.round(this.width);
        this.height = Math.round(this.height);
        this.#stale = true;
        return this;
    }
    multiply(x = 0, y = x, w = x, h = y) {
        return this.clone().multiplySelf(x, y, w, h);
    }
    multiplySelf(x = 0, y = x, w = x, h = y) {
        if (x instanceof Shape) {
            this.x *= x.x;
            this.y *= x.y;
            this.width *= x.width;
            this.height *= x.height;
        } else {
            this.x *= x;
            this.y *= y;
            this.width *= w;
            this.height *= h;
        }
        this.#stale = true;
        return this;
    }
    transform(matrix) {
        return this.clone().transformSelf(matrix);
    }
    transformSelf(matrix) {
        const { x , y , width , height  } = this;
        const { a , b , c , d , e , f  } = matrix;
        this.x = a * x + c * y + e;
        this.y = b * x + d * y + f;
        this.width = a * width + c * height;
        this.height = b * width + d * height;
        this.#stale = true;
        return this;
    }
    normalize(other) {
        return this.clone().normalizeSelf(other);
    }
    normalizeSelf(other) {
        if (other instanceof DOMMatrix) {
            this.transformSelf(other.inverse());
        } else {
            this.x -= other.left;
            this.y -= other.top;
            this.x /= other.width;
            this.y /= other.height;
            this.width /= other.width;
            this.height /= other.height;
        }
        this.#stale = true;
        return this;
    }
    equals(other) {
        return this.x === other.x && this.y === other.y && this.width === other.width && this.height === other.height;
    }
    intersectionRect(other) {
        return (this.bottom - 0.75 >= other.top ? Intersect.Top : 0) + (this.top + 0.75 <= other.bottom ? Intersect.Bottom : 0) + (this.right - 0.75 >= other.left ? Intersect.Left : 0) + (this.left + 0.75 <= other.right ? Intersect.Right : 0);
    }
    intersectsRect(other) {
        return !(this.bottom - 0.75 < other.top || this.top + 0.75 > other.bottom || this.right - 0.75 < other.left || this.left + 0.75 > other.right);
    }
    withinRect(other) {
        return this.left >= other.left && this.right <= other.right && this.top >= other.top && this.bottom <= other.bottom;
    }
    distanceRect(other) {
        let dx = 0;
        let dy = 0;
        if (this.right < other.left) {
            dx = other.left - this.right;
        } else if (this.left > other.right) {
            dx = this.left - other.right;
        } else {
            dx = Math.abs(this.center.x - other.center.x);
        }
        if (this.bottom < other.top) {
            dy = other.top - this.bottom;
        } else if (this.top > other.bottom) {
            dy = this.top - other.bottom;
        } else {
            dy = Math.abs(this.center.y - other.center.y);
        }
        return new Point(dx, dy);
    }
    collisionResponse(other) {
        return this.intersectPoint(other).addSelf(other.center).subSelf(this.center);
    }
    place(other, placement) {
        return this.clone().placeSelf(other, placement);
    }
    placeSelf(other, placement) {
        if (placement.includes('n')) this.y = other.top - this.height;
        else if (placement.includes('s')) this.y = other.bottom;
        else this.y = other.center.y - this.height * 0.5;
        if (placement.includes('w')) this.x = other.left - this.width;
        else if (placement.includes('e')) this.x = other.right;
        else this.x = other.center.x - this.width * 0.5;
        if (placement.includes('r')) {
            this.x += this.width;
        }
        if (placement.includes('l')) {
            this.x -= this.width;
        }
        this.#stale = true;
        return this;
    }
}

class Polygon {
    static toSVGPath(points) {
        return (points.length ? points : [
            [
                0,
                0
            ]
        ]).reduce((p, n)=>{
            let last = p.at(-1);
            if (last.length === 3) {
                last = [
                    'L'
                ];
                p.push(last);
            }
            last.push(...n);
            return p;
        }, [
            [
                'M'
            ]
        ]).flat(Infinity).join(' ');
    }
    static sum(points) {
        return points.reduce((p, n)=>{
            p.x += n.x;
            p.y += n.y;
            return p;
        }, new Point());
    }
    static rope(points, coeff = 1) {
        const from = points[0].clone();
        const to = points.at(-1).clone();
        const oldRope = Polyline.fromPoints(points);
        const mags = oldRope.lines.map((line)=>line.mag());
        const total = mags.reduce((a, b)=>a + b, 0);
        const normals = mags.map((mag)=>mag / total);
        const result = points.slice().map((x)=>x.clone());
        const pl = points.length;
        const pl_1 = pl - 1;
        const c = pl_1 / pl;
        let p = to.clone();
        for(let i = pl_1; i--;){
            const t = (i + 1) / pl_1;
            const np = result[i];
            const normal = normals[Math.ceil(i * c)];
            const d = p.screen(np).scale(normal).scale(t ** coeff);
            p.translateSelf(d.negate());
            np.translateSelf(d);
            p = np;
        }
        result[0] = from.clone();
        result[points.length - 1] = to.clone();
        return result;
    }
    static chop(points, min = 35, max = 240) {
        const from = points[0].clone();
        const to = points.at(-1).clone();
        const newRope = Polyline.fromPoints(points);
        const newPoints = [];
        let last;
        for (const line of newRope.lines){
            const mag = line.mag();
            if (mag > min) {
                last = newPoints.push(line.p1);
            } else {
                if (last) {
                    newPoints[last - 1].translateSelf(line.p1.screen(newPoints[last - 1]).scale(0.5));
                }
            }
            if (mag > max) {
                const d = line.p2.screen(line.p1);
                const np = line.p1.translate(d);
                last = newPoints.push(np);
            }
        }
        newPoints[0] = from.clone();
        newPoints.push(to.clone());
        return newPoints;
    }
    static morph(morphFn, from, to, t) {
        const [length, fc, tc] = Morph.coeffs(from, to);
        const fn = morphFn(from, to);
        return t < 1 ? Array.from({
            length
        }, (_, i)=>{
            const fi = Math.round(i * fc);
            const ti = Math.round(i * tc);
            return fn(fi, ti, t);
        }) : to;
    }
    static resample(points, index, t) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const p_0 = points[Math.min(index, pl_1)];
        const p_1 = points[Math.max(0, index - 1)];
        const p$1 = points[Math.min(index + 1, pl_1)];
        return p_0.interpolate(p_1.interpolate(p_0, t).translateSelf(p_0.interpolate(p$1, 1 - t)).scale(0.5), t);
    }
    static fit(points, length) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const coeff = pl / length;
        const newPoints = Array.from({
            length
        }, (_, i)=>{
            const index = Math.round(i * coeff);
            const p_0 = points[Math.min(index, pl_1)];
            const p_1 = points[Math.max(0, index - 1)];
            const p$1 = points[Math.min(index + 1, pl_1)];
            return p_0.interpolate(p_1.interpolate(p_0, 1).translateSelf(p_0.interpolate(p$1, 0)).scale(0.5), 0.5);
        });
        return newPoints;
    }
    // https://www.musicdsp.org/en/latest/Other/60-5-point-spline-interpollation.html
    static resampleSpline(points, index, t) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const p0 = points[Math.max(0, index - 2)];
        const p1 = points[Math.max(0, index - 1)];
        const p2 = points[Math.min(index, pl_1)];
        const p3 = points[Math.min(index + 1, pl_1)];
        const p4 = points[Math.min(index + 2, pl_1)];
        const p5 = points[Math.min(index + 3, pl_1)];
        const x = p2.x + 0.04166666666 * t * ((p3.x - p1.x) * 16.0 + (p0.x - p4.x) * 2.0 + t * ((p3.x + p1.x) * 16.0 - p0.x - p2.x * 30.0 - p4.x + t * (p3.x * 66.0 - p2.x * 70.0 - p4.x * 33.0 + p1.x * 39.0 + p5.x * 7.0 - p0.x * 9.0 + t * (p2.x * 126.0 - p3.x * 124.0 + p4.x * 61.0 - p1.x * 64.0 - p5.x * 12.0 + p0.x * 13.0 + t * ((p3.x - p2.x) * 50.0 + (p1.x - p4.x) * 25.0 + (p5.x - p0.x) * 5.0)))));
        // dprint-ignore
        const y = p2.y + 0.04166666666 * t * ((p3.y - p1.y) * 16.0 + (p0.y - p4.y) * 2.0 + t * ((p3.y + p1.y) * 16.0 - p0.y - p2.y * 30.0 - p4.y + t * (p3.y * 66.0 - p2.y * 70.0 - p4.y * 33.0 + p1.y * 39.0 + p5.y * 7.0 - p0.y * 9.0 + t * (p2.y * 126.0 - p3.y * 124.0 + p4.y * 61.0 - p1.y * 64.0 - p5.y * 12.0 + p0.y * 13.0 + t * ((p3.y - p2.y) * 50.0 + (p1.y - p4.y) * 25.0 + (p5.y - p0.y) * 5.0)))));
        return new Point(x, y);
    }
    // https://www.musicdsp.org/en/latest/Other/49-cubic-interpollation.html
    static resampleCubic(points, index, t) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const i = index + t | 0;
        const p_1 = points[Math.min(Math.max(0, i - 1), pl_1)];
        const p0 = points[Math.min(i, pl_1)];
        const p1 = points[Math.min(i + 1, pl_1)];
        const p2 = points[Math.min(i + 2, pl_1)];
        const ax = (3 * (p0.x - p1.x) - p_1.x + p2.x) * 0.5;
        const bx = 2 * p1.x + p_1.x - (5 * p0.x + p2.x) * 0.5;
        const cx = (p1.x - p_1.x) * 0.5;
        const x = ((ax * t + bx) * t + cx) * t + p0.x;
        const ay = (3 * (p0.y - p1.y) - p_1.y + p2.y) * 0.5;
        const by = 2 * p1.y + p_1.y - (5 * p0.y + p2.y) * 0.5;
        const cy = (p1.y - p_1.y) * 0.5;
        const y = ((ay * t + by) * t + cy) * t + p0.y;
        return new Point(x, y);
    }
    static boundingRect(points) {
        const minX = Math.min(...points.map((p)=>p.x));
        const minY = Math.min(...points.map((p)=>p.y));
        const maxX = Math.max(...points.map((p)=>p.x));
        const maxY = Math.max(...points.map((p)=>p.y));
        return new Rect(minX, minY, maxX - minX, maxY - minY);
    }
    // https://github.com/replit/kaboom/blob/d6985fde12b7073cf7bd300c93e1f79a93453a0a/src/math.ts#L867-L902
    static sat(p1, p2) {
        let overlap = Number.MAX_VALUE;
        let displacement = new Point();
        for (const poly of [
            p1,
            p2
        ]){
            for(let i = 0; i < poly.points.length; i++){
                const a = poly.points[i];
                const b = poly.points[(i + 1) % poly.points.length];
                const axisProj = b.sub(a).normal().unit();
                let min1 = Number.MAX_VALUE;
                let max1 = -Number.MAX_VALUE;
                for(let j = 0; j < p1.points.length; j++){
                    const q = p1.points[j].dot(axisProj);
                    min1 = Math.min(min1, q);
                    max1 = Math.max(max1, q);
                }
                let min2 = Number.MAX_VALUE;
                let max2 = -Number.MAX_VALUE;
                for(let j1 = 0; j1 < p2.points.length; j1++){
                    const q1 = p2.points[j1].dot(axisProj);
                    min2 = Math.min(min2, q1);
                    max2 = Math.max(max2, q1);
                }
                const o = Math.min(max1, max2) - Math.max(min1, min2);
                if (o < 0) {
                    return null;
                }
                if (o < Math.abs(overlap)) {
                    const o1 = max2 - min1;
                    const o2 = min2 - max1;
                    overlap = Math.abs(o1) < Math.abs(o2) ? o1 : o2;
                    displacement = axisProj.scale(overlap);
                }
            }
        }
        return displacement;
    }
    points;
    constructor(polygon = []){
        if (Array.isArray(polygon)) {
            this.points = polygon;
        } else {
            this.points = polygon.points;
        }
    }
}

class Morph {
    static coeffs = (from, to)=>{
        const fl = from.length;
        const tl = to.length;
        const length = Math.max(fl, tl);
        const fc = fl / length;
        const tc = tl / length;
        return [
            length,
            fc,
            tc,
            fl,
            tl
        ];
    };
    static Nearest = (from, to)=>{
        const fl_1 = from.length - 1;
        const tl_1 = to.length - 1;
        return (fi, ti, t)=>{
            const fp = from[Math.min(fi, fl_1)];
            const tp = to[Math.min(ti, tl_1)];
            return fp.interpolate(tp, t);
        };
    };
    static Linear = (from, to)=>(fi, ti, t)=>{
            const fp = Polygon.resample(from, fi, 0.5);
            const tp = Polygon.resample(to, ti, 0.5);
            return fp.interpolate(tp, t);
        };
    static Cubic = (from, to)=>(fi, ti, t)=>{
            const fp = Polygon.resampleCubic(from, fi, 0.5);
            const tp = Polygon.resampleCubic(to, ti, 0.5);
            return fp.interpolate(tp, t);
        };
    static Spline = (from, to)=>(fi, ti, t)=>{
            const fp = Polygon.resampleSpline(from, fi, 0.5);
            const tp = Polygon.resampleSpline(to, ti, 0.5);
            return fp.interpolate(tp, t);
        };
}

// Cardinal spline - a uniform Catmull-Rom spline with a tension option
const cardinal = (data, closed, tension)=>{
    if (data.length < 1) return 'M 0 0';
    if (tension == null) tension = 1;
    const size = data.length - (closed ? 0 : 1);
    const path = [
        `M ${data[0]} C`
    ];
    for(let i = 0; i < size; i++){
        let p0, p1, p2, p3;
        if (closed) {
            p0 = data[(i - 1 + size) % size];
            p1 = data[i];
            p2 = data[(i + 1) % size];
            p3 = data[(i + 2) % size];
        } else {
            p0 = i == 0 ? data[0] : data[i - 1];
            p1 = data[i];
            p2 = data[i + 1];
            p3 = i == size - 1 ? p2 : data[i + 2];
        }
        const x1 = p1.x + (p2.x - p0.x) / 6 * tension;
        const y1 = p1.y + (p2.y - p0.y) / 6 * tension;
        const x2 = p2.x - (p3.x - p1.x) / 6 * tension;
        const y2 = p2.y - (p3.y - p1.y) / 6 * tension;
        path.push(x1, y1, x2, y2, p2.x, p2.y);
    }
    closed && path.push('z');
    return path.join(' ');
};

const isMobileAgent = /Mobi|Android|Tablet/i.test(navigator.userAgent);

const observed = ()=>$.mixin((superclass)=>{
        return class ObservedMixin extends superclass {
            #offsetLeft;
            #offsetTop;
            #offsetWidth;
            #offsetHeight;
            get offsetLeft() {
                return this.#offsetLeft;
            }
            get offsetTop() {
                return this.#offsetTop;
            }
            get offsetWidth() {
                return this.#offsetWidth;
            }
            get offsetHeight() {
                return this.#offsetHeight;
            }
             #updateOffsets() {
                this.#offsetLeft = super.offsetLeft;
                // @ts-ignore
                this.#offsetTop = super.offsetTop;
                // @ts-ignore
                this.#offsetWidth = super.offsetWidth;
                // @ts-ignore
                this.#offsetHeight = super.offsetHeight;
            }
            rect = new $.Rect();
            pos = $(this).reduce(({ $ , rect  })=>$.pos?.equals(rect.pos) ? $.pos : rect.pos);
            size = $(this).reduce(({ $ , rect  })=>$.size?.equals(rect.size) ? $.size : rect.size);
            get ownRect() {
                this.#updateOffsets();
                const offset = $.getElementOffset(this);
                return new $.Rect(offset.x, offset.y, this.offsetWidth, this.offsetHeight);
            }
            mounted($) {
                $.rect = $.fulfill(({ host  })=>(fulfill)=>{
                        const update = ()=>fulfill(host.ownRect);
                        update();
                        $.observe.resize(host, update);
                    });
            }
        };
    });

var index = /*#__PURE__*/Object.freeze({
    __proto__: null,
    observed: observed
});

const observeNodes = (nodes, observer, options)=>nodes?.forEach((node)=>observer.observe(node, options));
const nodesToText = (nodes)=>nodes?.map((node)=>{
        const text = node.textContent;
        return text.trim().length ? text : '';
    }).join('') ?? '';
const slotted = (slots)=>({
        get nodes () {
            return slots.map((slot)=>slot.assignedNodes()).flat(Infinity);
        },
        get elements () {
            return slots.map((slot)=>slot.assignedElements()).flat(Infinity);
        },
        get firstChild () {
            return this.elements[0];
        }
    });
const onSlotChange = (el, cb, fn = (el)=>el instanceof HTMLSlotElement ? [
        el
    ] : [
        ...el.querySelectorAll('slot')
    ])=>on(el, 'slotchange')(()=>cb(slotted(fn(el))));
const onTextChange = (el, cb, fn)=>{
    let observer;
    const off = onSlotChange(el, ({ nodes  })=>{
        observer?.disconnect();
        if (nodes.length) {
            observer = new MutationObserver(()=>cb(nodesToText(nodes)));
            observeNodes(nodes, observer, {
                characterData: true
            });
            cb(nodesToText(nodes));
        }
    }, fn);
    return ()=>{
        observer?.disconnect();
        return off();
    };
};

const getElementOffset = (el)=>{
    let x = 0;
    let y = 0;
    do {
        x += el.offsetLeft - el.scrollLeft;
        y += el.offsetTop - el.scrollTop;
    }while (el = el.offsetParent)
    return {
        x,
        y
    };
};

const dataset = (target, data)=>{
    for(const key in data)target.dataset[key] = data[key].toString();
};

const findParentElement = (child, fn)=>{
    let el = child;
    while(el = el.parentElement)if (fn(el)) return el;
};

const getActiveElement = (el = document.activeElement)=>{
    if (el?.shadowRoot) return getActiveElement(el.shadowRoot.activeElement) ?? el;
    return el;
};

const kbd = (...shortcuts)=>(e)=>{
        const pressed = [
            e.shiftKey && 'shift',
            e.altKey && 'alt',
            e.ctrlKey && 'ctrl',
            e.metaKey && 'meta',
            (e.ctrlKey || e.metaKey) && 'cmd',
            e.key.replace('Arrow', '').toLowerCase(),
        ].filter(Boolean);
        for (const [shortcut, cmd] of shortcuts){
            if (shortcut.every((x)=>pressed.includes(x.toLowerCase()))) {
                e.preventDefault();
                e.stopPropagation();
                return cmd(e);
            }
        }
    };

const observe = {
    resize: (el, fn)=>{
        const observer = new ResizeObserver(fn);
        observer.observe(el);
        return ()=>observer.disconnect();
    },
    gc: (item, value, fn)=>{
        const reg = new FinalizationRegistry(fn);
        reg.register(item, value);
        return reg;
    }
};

const ondblclick = (el, cb, ms = 250)=>{
    let prevTimestamp = 0;
    const pointers = new Set();
    const clear = (e)=>{
        pointers.delete(e.pointerId);
    };
    return chain(on(el).pointerdown((e)=>{
        if (!(e.buttons & MouseButton.Left)) return;
        pointers.add(e.pointerId);
        if (pointers.size > 1) return;
        if (e.timeStamp - prevTimestamp < ms) {
            e.stopPropagation();
            cb();
            return;
        }
        prevTimestamp = e.timeStamp;
    }), on(window).pointerup(clear), on(window).pointercancel(clear));
};

const shadow = (host, init = '', html = '')=>{
    const root = host.attachShadow(typeof init === 'object' ? init : {
        mode: 'open'
    });
    root.innerHTML = typeof init === 'string' ? init : html;
    return root;
};

class TransitionSettings {
    immediate = bool;
    expire = bool;
    expireAfter = 0;
    locked = bool;
    pop = bool;
    drop = bool;
}
class Transition extends EventTarget {
    static from() {
        throw new TypeError('Transition cannot be recreated by its string representation.');
    }
    #modeExpireTime;
    #dropExpireTime;
    #modeTimeout;
    #resetModeTimeout;
    #locked;
    constructor(state, AnimSettings){
        super();
        this.state = state;
        this.AnimSettings = AnimSettings;
        this.#modeExpireTime = 0;
        this.#dropExpireTime = 0;
        this.#locked = false;
        this.to = toFluent(TransitionSettings, (settings)=>(mode, cb)=>{
                // TODO: this is bad design
                if (this.#locked) {
                    return;
                }
                const now = performance.now();
                if (now < this.#dropExpireTime) {
                    return;
                }
                clearTimeout(this.#resetModeTimeout);
                clearTimeout(this.#modeTimeout);
                if (now > this.#modeExpireTime || settings.immediate) {
                    this.#setMode(settings, mode, cb);
                } else {
                    this.#modeTimeout = setTimeout(()=>{
                        this.#setMode(settings, mode, cb);
                    }, this.#modeExpireTime - now);
                }
            });
        $.on(state).change(()=>{
            this.dispatchEvent(new CustomEvent('change'));
        });
    }
    get animSettings() {
        return this.AnimSettings[this.state.current] ?? this.AnimSettings[this.state.states.Idle];
    }
     #resetMode(cb) {
        clearTimeout(this.#resetModeTimeout);
        this.#resetModeTimeout = setTimeout(()=>{
            requestAnimationFrame(()=>{
                this.#dropExpireTime = 0;
                if (!this.state.isIdle) {
                    this.state.pop(this.state.current);
                }
                if (cb) requestAnimationFrame(cb);
            });
        }, this.AnimSettings[this.state.current]?.duration || 200);
    }
     #setMode(settings, mode, cb1) {
        const now = performance.now();
        if (mode === this.state.states.Idle) this.#resetMode(cb1);
        else {
            this.#modeExpireTime = now + (settings.expireAfter ? settings.expireAfter : settings.expire ? this.AnimSettings[mode]?.duration ?? 0 : 0);
            if (settings.drop) this.#dropExpireTime = this.#modeExpireTime;
            if (settings.locked) this.#locked = true;
            if (mode !== this.state.current) {
                requestAnimationFrame(()=>{
                    if (!this.state.isIdle) {
                        this.state.swap(mode);
                    } else {
                        this.state.push(mode);
                    }
                    if (cb1) requestAnimationFrame(cb1);
                    if (settings.pop) this.#resetMode();
                });
            } else {
                cb1?.();
                if (settings.pop) this.#resetMode();
            }
        }
    }
    unlock() {
        this.#locked = false;
        this.state.pop(this.state.current);
    }
    to;
    get current() {
        return this.toString();
    }
    toString() {
        return this.state.current;
    }
    state;
    AnimSettings;
}
const transition = (state, AnimSettings)=>new Transition(state, AnimSettings);

const warn = (fn)=>function(...args) {
        try {
            return fn.apply(this, args);
        } catch (e) {
            console.warn(e);
        }
    };

function _(ctx) {
    if (typeof ctx === 'function') {
        return ctx;
    } else {
        const c = ctx;
        const shadow = $.shadow.bind(null, ctx);
        const slotted = slotted$1(c);
        return Getter((key)=>{
            switch(key){
                case 'shadow':
                    return shadow;
                case 'slotted':
                    return slotted;
                case 'state':
                    return (states)=>new State(states);
                default:
                    {
                        return FluentCapture()[key];
                    }
            }
        });
    }
}
const mixin = (mixin)=>(realsuperclass)=>element()(mixin(realsuperclass));

var sigl = /*#__PURE__*/Object.freeze({
    __proto__: null,
    ContextClass: ContextClass,
    element: element,
    render: render,
    State: State,
    _: _,
    mixin: mixin,
    mix: pipeInto,
    isMobile: isMobileAgent,
    mixins: index,
    AbortOptions: AbortOptions,
    abort: abort,
    MouseButton: MouseButton,
    DispatchOptions: DispatchOptions,
    dispatch: dispatch,
    dispatchBind: dispatchBind,
    EventOptions: EventOptions,
    wrapEvent: wrapEvent,
    event: event,
    chain: chain,
    onAll: onAll,
    OnOptions: OnOptions,
    on: on,
    QueueOptions: QueueOptions,
    wrapQueue: wrapQueue,
    queue: queue,
    Task: Task,
    runTask: runTask,
    groupTasks: groupTasks,
    css: css,
    cssToJs: cssToJs,
    jsToCss: jsToCss,
    get Intersect () { return Intersect; },
    Line: Line,
    Matrix: Matrix,
    Morph: Morph,
    Point: Point,
    Polygon: Polygon,
    Polyline: Polyline,
    Rect: Rect,
    Scalar: Scalar,
    Shape: Shape,
    cardinal: cardinal,
    PropertySettings: PropertySettings,
    protoPropertyMap: protoPropertyMap,
    prop: prop,
    attr: attr,
    out: out,
    is: is,
    RefProxy: RefProxy,
    Ref: Ref,
    PropsRef: PropsRef,
    RefCollection: RefCollection,
    RefMap: RefMap,
    RefSet: RefSet,
    observeNodes: observeNodes,
    nodesToText: nodesToText,
    slotted: slotted,
    onSlotChange: onSlotChange,
    onTextChange: onTextChange,
    getElementOffset: getElementOffset,
    dataset: dataset,
    findParentElement: findParentElement,
    getActiveElement: getActiveElement,
    kbd: kbd,
    observe: observe,
    ondblclick: ondblclick,
    shadow: shadow,
    TransitionSettings: TransitionSettings,
    Transition: Transition,
    transition: transition,
    warn: warn
});

// @ts-ignore
globalThis.Reflect.metadata ??= ()=>{};
const $ = Object.assign(_, sigl, {
    String: String,
    Number: Number,
    Boolean: Boolean
});

const getViewportRect = () => new Rect(document.scrollingElement.scrollLeft, document.scrollingElement.scrollTop, window.visualViewport.width, window.visualViewport.height);
// TODO: @$.reactive()
class PopupScene {
    surface;
    viewportRect = getViewportRect();
    viewMatrix = new Matrix();
    runCollisions;
    // @ts-ignore
    $;
    context;
    popups = new Set();
    constructor(surface) {
        this.surface = surface;
        $.ContextClass.attach(this, $);
        this.create();
    }
    destroy() {
        this.context.cleanup();
    }
    create() {
        const { $ } = this;
        // read surface
        $.effect(({ surface }) => surface.$.effect(({ viewMatrix }) => {
            $.viewMatrix = new Matrix([...viewMatrix.toFloat64Array()]);
        }));
        $.effect(({ surface }) => $.observe.resize(surface, () => {
            $.viewportRect = Rect.fromElement(surface); //.getBoundingClientRect()) //getViewportRect()
        }));
        $.runCollisions = $.reduce(({ popups }) => (() => {
            for (const popup of popups) {
                popup.rect = popup.place(popup.placement ?? popup.originalPlacement)
                    .containSelf($.viewportRect);
            }
            for (let i = 0; i < 10; i++) {
                for (const a of popups) {
                    for (const b of popups) {
                        if (a === b)
                            continue;
                        if (a.rect.intersectsRect(b.rect)) {
                            const c = a.rect.collisionResponse(b.rect);
                            a.rect.translateSelf(c.scale(0.5)).containSelf($.viewportRect);
                            b.rect.translateSelf(c.scale(-0.5)).containSelf($.viewportRect);
                        }
                    }
                    a.rect = a.rect.clone();
                }
            }
        }));
        $.effect(({ surface, runCollisions }) => surface.$.effect(({ viewStyleTransform: _ }) => {
            runCollisions();
        }));
    }
}

/******************************************************************************
Copyright (c) Microsoft Corporation.
Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.
THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
function __decorate$6(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
}
function __metadata(metadataKey, metadataValue) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(metadataKey, metadataValue);
}

const Easing = {
    Flat: [
        1,
        1,
        1,
        1
    ],
    Linear: [
        0,
        0,
        1,
        1
    ]
};

// npm: bezier-easing
// vendored here because of esm issues
var NEWTON_ITERATIONS = 4;
var NEWTON_MIN_SLOPE = 1e-3;
var SUBDIVISION_PRECISION = 1e-7;
var SUBDIVISION_MAX_ITERATIONS = 10;
var kSplineTableSize = 11;
var kSampleStepSize = 1 / (kSplineTableSize - 1);
var float32ArraySupported = typeof Float32Array === 'function';
function A(aA1, aA2) {
    return 1 - 3 * aA2 + 3 * aA1;
}
function B(aA1, aA2) {
    return 3 * aA2 - 6 * aA1;
}
function C(aA1) {
    return 3 * aA1;
}
function calcBezier(aT, aA1, aA2) {
    return ((A(aA1, aA2) * aT + B(aA1, aA2)) * aT + C(aA1)) * aT;
}
function getSlope(aT, aA1, aA2) {
    return 3 * A(aA1, aA2) * aT * aT + 2 * B(aA1, aA2) * aT + C(aA1);
}
function binarySubdivide(aX, aA, aB, mX1, mX2) {
    var currentX, currentT, i = 0;
    do {
        currentT = aA + (aB - aA) / 2;
        currentX = calcBezier(currentT, mX1, mX2) - aX;
        if (currentX > 0) {
            aB = currentT;
        } else {
            aA = currentT;
        }
    }while (Math.abs(currentX) > SUBDIVISION_PRECISION && ++i < SUBDIVISION_MAX_ITERATIONS)
    return currentT;
}
function newtonRaphsonIterate(aX, aGuessT, mX1, mX2) {
    for(var i = 0; i < NEWTON_ITERATIONS; ++i){
        var currentSlope = getSlope(aGuessT, mX1, mX2);
        if (currentSlope === 0) {
            return aGuessT;
        }
        var currentX = calcBezier(aGuessT, mX1, mX2) - aX;
        aGuessT -= currentX / currentSlope;
    }
    return aGuessT;
}
function LinearEasing(x) {
    return x;
}
var src = function bezier(mX1, mY1, mX2, mY2) {
    if (!(0 <= mX1 && mX1 <= 1 && 0 <= mX2 && mX2 <= 1)) {
        throw new Error('bezier x values must be in [0, 1] range');
    }
    if (mX1 === mY1 && mX2 === mY2) {
        return LinearEasing;
    }
    var sampleValues = float32ArraySupported ? new Float32Array(kSplineTableSize) : new Array(kSplineTableSize);
    for(var i = 0; i < kSplineTableSize; ++i){
        sampleValues[i] = calcBezier(i * kSampleStepSize, mX1, mX2);
    }
    function getTForX(aX) {
        var intervalStart = 0;
        var currentSample = 1;
        var lastSample = kSplineTableSize - 1;
        for(; currentSample !== lastSample && sampleValues[currentSample] <= aX; ++currentSample){
            intervalStart += kSampleStepSize;
        }
        --currentSample;
        var dist = (aX - sampleValues[currentSample]) / (sampleValues[currentSample + 1] - sampleValues[currentSample]);
        var guessForT = intervalStart + dist * kSampleStepSize;
        var initialSlope = getSlope(guessForT, mX1, mX2);
        if (initialSlope >= NEWTON_MIN_SLOPE) {
            return newtonRaphsonIterate(aX, guessForT, mX1, mX2);
        } else if (initialSlope === 0) {
            return guessForT;
        } else {
            return binarySubdivide(aX, intervalStart, intervalStart + kSampleStepSize, mX1, mX2);
        }
    }
    return function BezierEasing(x) {
        if (x === 0) {
            return 0;
        }
        if (x === 1) {
            return 1;
        }
        return calcBezier(getTForX(x), mY1, mY2);
    };
};

const createStepAnimation = ({ duration , easing  }, anim)=>{
    let from;
    let to;
    let last;
    let current;
    let state = 'finished';
    const easingFn = src(...easing);
    if (anim) {
        current = from = anim.current;
        state = 'preparing';
    }
    let startTime = 0;
    let now;
    const timeElapsed = ()=>{
        now = performance.now();
        return now - startTime;
    };
    const set = (values)=>{
        switch(state){
            case 'finished':
                if (from) {
                    current = from;
                } else {
                    current = from = values;
                    last = to = values;
                    state = 'preparing';
                    return current;
                }
            case 'preparing':
                startTime = performance.now();
                from = current;
                last = to = values;
                state = 'running';
                return {
                    ...current
                };
            case 'running':
                last = values;
                return current;
        }
    };
    const update = (values)=>{
        switch(state){
            case 'running':
                current = values;
                break;
            case 'finished':
                current = from = last;
                break;
        }
        return current;
    };
    return {
        get t () {
            if (state === 'running') {
                const elapsed = timeElapsed();
                let t = elapsed / duration;
                if (t < 1) return easingFn(t);
                // schedule next
                if (last !== to) {
                    // startTime = performance.now()
                    startTime += duration;
                    t = (elapsed - duration) / duration;
                    from = current;
                    to = last;
                    if (t < 1) return easingFn(t);
                }
                state = 'finished';
            }
            return 1;
        },
        get from () {
            return from;
        },
        get to () {
            return to;
        },
        get state () {
            return state;
        },
        get current () {
            return current;
        },
        set,
        update
    };
};

const round = ({ x , y  }, p = 1)=>({
        x: Math.round(x * p) / p,
        y: Math.round(y * p) / p
    });
const modwrap = (x, N)=>(x % N + N) % N;

var __decorate$5 = undefined && undefined.__decorate || function(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for(var i = decorators.length - 1; i >= 0; i--)if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
const style$3 =  `
[part=canvas] {
  contain: size layout style paint;
  position: absolute;
  left: 0;
  top: 0;
  width: 100%;
  height: 100%;
  display: inline;
  overflow: hidden;
}`;
let SurfaceGridElement = class SurfaceGridElement extends HTMLElement {
    root = $.shadow(this, /*html*/ `<style>${style$3}</style><canvas part="canvas"></canvas>`);
    pixelRatio = window.devicePixelRatio;
    cellSize = 80;
    xPattern = '1' // 01010'
    ;
    yPattern = '1' // 01010'
    ;
    // from surface
    matrix;
    rect;
    pointers = new Map();
    slotted;
    transition;
    // animSettings?: SurfaceAnimSettings
    canvas;
    ctx;
    anim;
    animValues;
    zoom;
    offsetX;
    offsetY;
    mounted($) {
        $.canvas = $.query('[part=canvas]');
        $.ctx = $.reduce(({ canvas: grid  })=>grid.getContext('2d', {
                alpha: false,
                desynchronized: true
            }));
        $.effect(({ rect , canvas , pixelRatio  })=>{
            const size = rect.size.scaleSelf(pixelRatio).roundSelf();
            if (!size.equals(Point.fromObject(canvas))) Object.assign(canvas, size.toSizeObject());
        });
        $.anim = $.reduce.throttle(25).first.last.next(({ transition  })=>createStepAnimation(transition.animSettings, $.anim));
        $.animValues = $.reduce(({ anim: { set  } , matrix , pixelRatio  })=>set({
                zoom: matrix.a,
                offset: Point.fromMatrix(matrix).scaleSelf(pixelRatio)
            }));
        $.animValues = $.reduce.raf(({ anim: { t , from , to , update  } , animValues: _  })=>update({
                zoom: from.zoom + (to.zoom - from.zoom) * t,
                offset: from.offset.translate(to.offset.screen(from.offset).scale(t))
            }));
        // convert to scalar so they can be compared during
        // the canvas render effect so there are no wasteful redraws
        $.effect(({ animValues  })=>{
            $.zoom = animValues.zoom;
            $.offsetX = animValues.offset.x;
            $.offsetY = animValues.offset.y;
        });
        $.effect.raf(({ canvas: { width: w , height: h ,  } , rect: _ , ctx , cellSize , xPattern: xp , yPattern: yp , pixelRatio , zoom , offsetX , offsetY ,  })=>{
            ctx.fillStyle = '#222';
            ctx.fillRect(0, 0, w, h);
            ctx.strokeStyle = '#eee';
            ctx.lineWidth = Math.min(1, zoom ** 1.25 * 0.28);
            if (ctx.lineWidth < 0.01) return;
            const z = zoom * cellSize * pixelRatio;
            ctx.beginPath();
            let i = 0;
            let sx = offsetX;
            while(sx > 0)sx -= z, i--;
            while(sx < -z)sx += z, i++;
            for(let x = sx; x < w; x += z){
                if (!+xp[modwrap(i++, xp.length)]) continue;
                ctx.moveTo(x, 0);
                ctx.lineTo(x, h);
            }
            i = 0;
            let sy = offsetY;
            while(sy > 0)sy -= z, i--;
            while(sy < -z)sy += z, i++;
            for(let y = sy; y < h; y += z){
                if (!+yp[modwrap(i++, yp.length)]) continue;
                ctx.moveTo(0, y);
                ctx.lineTo(w, y);
            }
            ctx.stroke();
        });
    }
};
__decorate$5([
    $.attr()
], SurfaceGridElement.prototype, "pixelRatio", void 0);
__decorate$5([
    $.attr()
], SurfaceGridElement.prototype, "cellSize", void 0);
__decorate$5([
    $.attr()
], SurfaceGridElement.prototype, "xPattern", void 0);
__decorate$5([
    $.attr()
], SurfaceGridElement.prototype, "yPattern", void 0);
SurfaceGridElement = __decorate$5([
    $.element()
], SurfaceGridElement);

var __decorate$4 = undefined && undefined.__decorate || function(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for(var i = decorators.length - 1; i >= 0; i--)if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
const style$2 = /*css*/ `
[part=canvas] {
  contain: size layout style paint;
  position: fixed;
  opacity: 0.6;
  width: 150px;
  height: 112.5px;
  right: 30px;
  bottom: 30px;
  z-index: 2;
  border-radius: 3px;

  /* this is cool but glitches on mobile */
  /* ${$.isMobile ? '' : 'transition: opacity 0.08s ease-out;'} */
}

[part=canvas].hide {
  opacity: 0;
  pointer-events: none;
}

[part=canvas]:hover {
  opacity: 0.9;
}`;
let SurfaceMinimapElement = class SurfaceMinimapElement extends HTMLElement {
    root = $(this).shadow(/*html*/ `<style>${style$2}</style><canvas part="canvas" class="hide"></canvas>`);
    pixelRatio = window.devicePixelRatio;
    scale = 0.2;
    ratio = 4 / 3;
    // from surface
    matrix;
    rect;
    items;
    canvas = $(this).query( `[part=canvas]`);
    ctx = $(this).reduce(({ canvas  })=>canvas.getContext('2d', {
            alpha: false,
            desynchronized: true
        })) //?: CanvasRenderingContext2D
    ;
    isOverMinimap = $(this).fulfill(({ canvas  })=>(fulfill)=>$.chain($.on(canvas).pointermove(()=>fulfill(true)), $.on(canvas).pointerleave(()=>fulfill(false))), false);
    allVisible = true;
    mounted($) {
        $.effect(({ rect , canvas , scale , ratio , pixelRatio  })=>{
            const [w, h] = rect.size.scaleSelf(pixelRatio);
            const g = Math.max(850, Math.min(1150 + w * 0.1, w * 0.15 + h * 0.85));
            const size = new Point(g * scale * ratio, g * scale);
            if (!size.round().equals(Point.fromObject(canvas))) {
                Object.assign(canvas, size.round().toSizeObject());
                Object.assign(canvas.style, size.normalize(pixelRatio).toStyleSize());
            }
        });
        $.effect.raf(({ ctx: ctx , canvas: { width: w , height: h  } , matrix , items: surfaceItems  })=>$.use(({ isOverMinimap , rect  })=>{
                if (!rect) return;
                const frame = rect.clone();
                const items = filterMap(surfaceItems, (el)=>!el.points && el.rect.transform(matrix).translateSelf(frame));
                const all = [
                    frame,
                    ...items
                ];
                const screen = Rect.combine(all);
                all.forEach((b)=>b.normalizeSelf(screen).multiplySelf(w, h));
                if (!isOverMinimap) {
                    $.allVisible = items.every((b)=>b.withinRect(frame));
                    if ($.allVisible) return;
                }
                ctx.clearRect(0, 0, w, h);
                ctx.fillStyle = '#282828';
                ctx.fillRect(0, 0, w, h);
                const lw = ctx.lineWidth;
                const lw2 = lw * 2;
                ctx.fillStyle = 'rgba(255,255,255,.045)';
                ctx.strokeStyle = 'rgba(255,255,255,.5)';
                ctx.beginPath();
                // @ts-ignore
                ctx.roundRect(frame.x + lw, frame.y + lw, frame.width - lw2, frame.height - lw2, [
                    3
                ]);
                ctx.fill();
                ctx.stroke();
                // draw items
                ctx.fillStyle = 'rgba(255,255,255,.16)';
                ctx.strokeStyle = '#aaa';
                for (const item of items){
                    const { x , y , width , height  } = item;
                    ctx.beginPath();
                    // @ts-ignore
                    ctx.roundRect(x + lw, y + lw, width - lw2, height - lw2, [
                        0.75
                    ]);
                    ctx.fill();
                    ctx.stroke();
                }
                // draw points/paths
                const paths = surfaceItems.filter((x)=>x.points).map((p)=>p.points.map((x)=>x.transform(matrix).normalizeSelf(screen.size).multiplySelf(w, h).translateSelf(frame)));
                ctx.strokeStyle = '#666';
                ctx.lineWidth = 1.75;
                for (const points of paths){
                    ctx.beginPath();
                    for (const p of points)ctx.lineTo(p.x, p.y);
                    ctx.stroke();
                }
            }));
        $.effect.raf(({ allVisible , canvas  })=>{
            if (allVisible && !canvas.classList.contains('hide')) canvas.classList.add('hide');
            if (!allVisible && canvas.classList.contains('hide')) canvas.classList.remove('hide');
        });
    }
};
__decorate$4([
    $.attr()
], SurfaceMinimapElement.prototype, "pixelRatio", void 0);
__decorate$4([
    $.attr()
], SurfaceMinimapElement.prototype, "scale", void 0);
__decorate$4([
    $.attr()
], SurfaceMinimapElement.prototype, "ratio", void 0);
SurfaceMinimapElement = __decorate$4([
    $.element()
], SurfaceMinimapElement);

var __decorate$3 = undefined && undefined.__decorate || function(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for(var i = decorators.length - 1; i >= 0; i--)if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
const SurfaceState = {
    Idle: 'surfaceidle',
    Overlay: 'surfaceoverlay',
    CenteringItem: 'surfacecenteringitem',
    CenteringView: 'surfacecenteringview',
    Connecting: 'surfaceconnecting',
    FullSize: 'surfacefullsize',
    MinimapPanning: 'surfaceminimappanning',
    Panning: 'surfacepanning',
    Pinching: 'surfacepinching',
    Selecting: 'surfaceselecting',
    Wheeling: 'surfacewheeling'
};
const SurfaceCursorState = {
    Idle: 'default',
    Copy: 'copy',
    Grabbing: 'grabbing',
    EWResize: 'ew-resize',
    NSResize: 'ns-resize',
    NWSEResize: 'nwse-resize'
};
const Idle = $.isMobile ? {
    duration: 0,
    easing: Easing.Flat
} : {
    duration: 0,
    easing: Easing.Flat
};
const Panning = $.isMobile ? {
    duration: 120,
    easing: [
        0,
        0,
        1,
        1
    ]
} : {
    duration: 150,
    easing: [
        0,
        0,
        1,
        1
    ]
};
const SurfaceAnimSettings = {
    [SurfaceState.Idle]: Idle,
    [SurfaceState.Overlay]: Panning,
    [SurfaceState.Selecting]: Idle,
    [SurfaceState.FullSize]: Idle,
    [SurfaceState.Panning]: Panning,
    [SurfaceState.MinimapPanning]: Panning,
    [SurfaceState.Pinching]: $.isMobile ? {
        duration: 200,
        easing: [
            0,
            0.15,
            0.45,
            1
        ]
    } : {
        duration: 200,
        easing: [
            0,
            0.55,
            0.25,
            1
        ]
    },
    [SurfaceState.Wheeling]: $.isMobile ? {
        duration: 500,
        easing: [
            0,
            0.55,
            0.25,
            1
        ]
    } : {
        duration: 185,
        easing: [
            0.1,
            0.15,
            0.75,
            1
        ]
    },
    // : {
    //   duration: 100,
    //   easing: [0, 0.25, 0.75, 1],
    // [SurfaceState.ItemMove]: ($.isMobile
    //     duration: 200,
    //     easing: [0, 0.15, 0.45, 1],
    //   : {
    //     duration: 150,
    //     easing: [0, 0, 0.2, 1],
    //////
    [SurfaceState.CenteringItem]: $.isMobile ? {
        duration: 600,
        easing: [
            0,
            0.67,
            0.045,
            1
        ]
    } : {
        duration: 415,
        easing: [
            0,
            0.12,
            0.29,
            1
        ]
    },
    // : {
    //   duration: 280, // 280
    //   easing: [0, 0.55, 0.25, 1], // [0, 0.55, 0.25, 1],
    [SurfaceState.CenteringView]: $.isMobile ? {
        duration: 400,
        easing: [
            0,
            0.2,
            0.001,
            1
        ]
    } : {
        duration: 415,
        easing: [
            0,
            0.1,
            0.05,
            1
        ]
    }
};
const style$1 = /*css*/ `

:host {
  contain: size layout style paint;
  position: relative;
  user-select: none;
  touch-action: none;
  display: flex;
  box-sizing: border-box;
  width: 100%;
  height: 100%;
  overflow: hidden;
}

[part=overlay] {
  position: fixed;
  left: 0;
  top: 0;
  right: 0;
  bottom: 0;
  opacity: 0;
  z-index: -1;
}

${[
    SurfaceState.FullSize,
    SurfaceState.Connecting,
    SurfaceState.MinimapPanning,
    SurfaceState.Overlay,
    SurfaceState.Panning,
    SurfaceState.Pinching,
    SurfaceState.Selecting,
].map((x)=>/*css*/ `:host([state=${x}]) [part=overlay]`)} {
  z-index: 100000;
}

${Object.entries(SurfaceAnimSettings).map(([name, { duration , easing  }])=>/*css*/ `
    :host([transition=${name}]) [part=view] {
      transition: transform ${duration}ms cubic-bezier(${easing});
    }
  `).join('')}

slot {
  contain: size layout style paint;
}

::slotted(*) {
  contain: size layout style;
  position: absolute;
}

[part=frame] {
  contain: size layout style paint;
  position: absolute;
  left: 0;
  top: 0;
  right: 0;
  bottom: 0;
  overflow: hidden;
}
[part=view] {
  contain: size layout style;
  transform-origin: 0 0 0;
  position: absolute;
  display: inline;
  z-index: 1;
}
[part=minimap] {
  display: none;
  contain: size style;
  z-index: 100001;
}

[part=full-size] {
  position: absolute;
  display: none;
  z-index: 100001; /* 1 over overlay */
  left: 0;
  top: 0;
  right: 0;
  bottom: 0;
  background: #000;
  > * {
    position: absolute;
    left: 0 !important;
    top: 0 !important;
    right: 0 !important;
    bottom: 0 !important;
    width: 100% !important;
    height: 100% !important;
    resize: none !important;
  }
}
:host([state=${SurfaceState.FullSize}]) [part=full-size] {
  display: block;
}

[part=selection] {
  position: absolute;
  background: #44b5;
  border: 1.5px solid #77a;
  pointer-events: none;
  box-sizing: border-box;
  border-radius: 1.3px;
  z-index: 1000;
  opacity: 0;
}
:host([state=${SurfaceState.Selecting}]) [part=selection] {
  opacity: 1;
  transition: ${[
    'left',
    'top',
    'width',
    'height'
].map((x)=>`${x} 9ms cubic-bezier(0.5,0,0.95,1)`)};
}
`;
let SurfaceElement = class SurfaceElement extends $.mix(HTMLElement, $.mixins.observed()) {
    static zIndex = 0;
    // dependencies
    Minimap = $.element(SurfaceMinimapElement);
    Grid = $.element(SurfaceGridElement);
    root = $(this).shadow();
    // attributes
    state = $(this).state(SurfaceState);
    cursorState = $(this).state(SurfaceCursorState);
    transition = $.transition(this.state, SurfaceAnimSettings);
    pixelRatio = window.devicePixelRatio;
    gridCellSize = 80;
    snapThreshold = 0.15;
    minZoom = 0.05
    ;
    maxZoom = 1.25;
    // sent to minimap
    minimapScale = 0.2;
    minimapRatio = 4 / 3;
    xPattern = '1' // 01010'
    ;
    yPattern = '1' // 01010'
    ;
    // minimap
    minimap;
    view;
    viewFrame;
    viewFrameRect;
    viewFrameNormalRect;
    viewRect;
    viewMatrix = new DOMMatrix().scale(0.001);
    viewMatrixString;
    viewStyleTransform;
    /** Holds the last centered item. */ centeredItem;
    /** Centers a specific rectangle in view. */ centerRect;
     centerItem;
    /** Centers item n+-diff sorted horizontally. */ centerOtherItem;
     centerView;
    /** Get the center matrix for any given rectangle. */ getCenterMatrix;
    /** Remember what we centered last, so Escape does the opposite */ didCenterLast = 'view';
    items = $(this).slotted.deep((el)=>(el.surface = this, el));
    hoveringItem;
    // pointers
    pointer;
    pointers = new Map();
    pinchStartMatrix;
    pinchStartDistance = 1;
    getPointerPos;
    // selecting
    selection;
    selectingRect;
    selectingStartPos;
    // full-size
    fullSize;
    exitFullSize;
    makeFullSize = $(this).reduce(({ $ , fullSize  })=>(item)=>{
            const prevParent = item.parentElement;
            const prevNext = item.nextSibling;
            item.fullSize = true;
            fullSize.appendChild(item);
            $.transition.to.immediate.locked(SurfaceState.FullSize);
            item.blur();
            requestAnimationFrame(()=>{
                item.focus();
            });
            $.exitFullSize = ()=>{
                offKeyDownFull();
                prevParent.insertBefore(item, prevNext);
                item.fullSize = false;
                $.transition.unlock();
                item.blur();
                requestAnimationFrame(()=>{
                    item.focus();
                });
                $.exitFullSize = null;
            };
            const offKeyDownFull = $.on(window).keydown.capture((e)=>{
                if (e.key === 'Escape') {
                    e.preventDefault();
                    e.stopImmediatePropagation();
                    $.exitFullSize();
                }
            });
        });
    setViewStyleTransform;
    get matrix() {
        return this.viewMatrix;
    }
    mounted($) {
        //!time 'state'
        //!time 'animating'
        $.effect(({ state: _  })=>{
            // console.log(_.current)
            //!? 'state', _.current
            return;
        });
        // host
        // $.effect(({ state }) => {
        //   document.querySelector('#debug')!.textContent = state
        $.effect(({ state  })=>{
            if ($.cursorState.isIdle) {
                if (state.is(SurfaceState.Panning)) {
                    $.cursorState.push(SurfaceCursorState.Grabbing);
                }
            } else if (state.is(SurfaceState.Idle)) {
                $.cursorState.pop($.cursorState.current);
            }
        });
        $.effect(({ host , cursorState  })=>{
            host.style.cursor = cursorState.current;
        });
        $.effect.raf(({ host , viewMatrix: _  })=>{
            $.dispatch(host, 'scroll');
        });
        $.effect.raf(({ host , rect: _  })=>{
            $.dispatch(host, 'resize');
        });
        $.effect(()=>$.chain($.on(window).keydown((e)=>{
                if (e.altKey && $.state.isIdle) {
                    $.state.push(SurfaceState.Overlay);
                }
            }), $.on(window).keyup((e)=>{
                if (!e.altKey && $.state.is(SurfaceState.Overlay)) {
                    $.state.pop(SurfaceState.Overlay);
                }
            })));
        // const refreshItems = $.queue.debounce(1000)(() => {
        // })
        // $.effect(({ host }) =>
        //   $.on(host).change(e => {
        //     const [origin] = e.composedPath()
        //     if (origin instanceof SurfaceMoveElement || origin instanceof SurfaceResizeElement) {
        //     }
        // )
        $.effect(({ items  })=>$.chain([
                ...items
            ].flatMap((x)=>[
                    $.on(x).pointerenter(()=>{
                        $.hoveringItem = x;
                    }),
                    $.on(x).pointerleave(()=>{
                        if ($.hoveringItem === x) {
                            $.hoveringItem = null;
                        }
                    }),
                ])));
        $.effect(({ hoveringItem  })=>{
            $.centeredItem = hoveringItem;
        });
        // view
        $.viewRect = $.fulfill(({ items  })=>(fulfill)=>{
                const refresh = $.queue.debounce(50)(()=>{
                    fulfill(Rect.combine(items.map((x)=>x.rect)));
                });
                return $.chain(items.map((item)=>$.chain(item.$.effect(({ rect: _  })=>{
                        refresh();
                    }), $.on(item).focus($.atomic(()=>{
                        // $.didCenterLast = 'view'
                        $.centeredItem = item;
                    })))));
            });
        $.viewFrameRect = $.reduce(({ rect: _ , viewFrame  })=>Rect.fromElement(viewFrame));
        $.viewFrameNormalRect = $.reduce(({ viewFrameRect , viewMatrix  })=>viewFrameRect.normalize(viewMatrix));
        // centering view and items
        $.getCenterMatrix = $.reduce(({ viewFrameRect  })=>(targetRect, paddingPct = 0)=>{
                const padding = Math.max(...targetRect.size.scale(paddingPct));
                const scaled = targetRect.scaleLinear(padding);
                const scale = Math.min(...viewFrameRect.size.normalize(scaled.size), 1);
                const target = new DOMMatrix().translateSelf(...scaled.center.negateSelf().translateSelf(padding * 0.5) //, padding * (paddingPct < 0.2 ? 0.7 : 0.6)) // breathe more at top because of the labels
                .scaleSelf(scale).translateSelf(viewFrameRect.size.scaleSelf(0.5)).pos).scaleSelf(scale);
                return target;
            });
        $.centerRect = $.reduce(({ getCenterMatrix  })=>$.atomic((rect)=>{
                $.didCenterLast = 'view';
                $.transition.to.pop.expire(SurfaceState.CenteringView, ()=>{
                    $.viewMatrix = getCenterMatrix(rect, 0.12);
                });
            }));
        $.centerItem = $.reduce(({ getCenterMatrix  })=>$.atomic((item)=>{
                $.centeredItem = item;
                $.didCenterLast = 'item';
                if (!$.isMobile) {
                    item.blur();
                    item.focus({
                        preventScroll: true
                    });
                }
                $.transition.to.pop.expire(SurfaceState.CenteringItem, ()=>{
                    $.viewMatrix = getCenterMatrix(item.rect, 0.075);
                });
            }));
        $.centerOtherItem = $.reduce(({ getCenterMatrix , items  })=>$.callback(({ centeredItem  })=>(diff)=>{
                    const sorted = filterMap(items, (el)=>!el.points && [
                            el,
                            el.rect
                        ]).sort(([, a], [, b])=>a.x === b.x ? a.y < b.y ? -1 : 1 : a.x < b.x ? -1 : 1);
                    let index = sorted.findIndex(([el])=>centeredItem === el);
                    if (!~index && diff < 0) index = 0;
                    const [item, itemRect] = sorted[modwrap(index + diff, sorted.length)];
                    $.centeredItem = item;
                    $.didCenterLast = 'item';
                    if (!$.isMobile) {
                        item.blur();
                        item.focus({
                            preventScroll: true
                        });
                    }
                    $.transition.to.pop.expire(SurfaceState.CenteringItem, ()=>{
                        $.viewMatrix = getCenterMatrix(itemRect, 0.35);
                    });
                }));
        // $.effect.raf(({ centeredItem }) => centeredItem.focus({ preventScroll: true }))
        $.centerView = $.reduce(({ getCenterMatrix  })=>$.atomic((mode = SurfaceState.CenteringView, paddingPct = 0.11)=>{
                $.didCenterLast = 'view';
                requestAnimationFrame(()=>{
                    $.cursorState.returnToIdle();
                });
                $.transition.to.immediate.pop.expire.drop(mode, ()=>{
                    if (!$.viewRect) return;
                    $.viewMatrix = getCenterMatrix($.viewRect, paddingPct);
                });
            }));
        $.effect.once(({ centerView  })=>{
            centerView(SurfaceState.Idle, 0.25);
        });
        $.effect(({ host , centerView  })=>$.ondblclick(host, centerView));
        $.effect(({ centerView , centerItem  })=>$.on(window).keydown($.kbd([
                [
                    'Escape'
                ],
                ()=>{
                    if ($.centeredItem && $.didCenterLast === 'view') {
                        centerItem($.centeredItem);
                    } else {
                        centerView();
                    }
                }
            ])));
        $.effect(({ centerOtherItem  })=>$.on(window).keydown.throttle(50)($.kbd([
                [
                    'Alt',
                    'Right'
                ],
                ()=>centerOtherItem(+1)
            ], [
                [
                    'Alt',
                    'Left'
                ],
                ()=>centerOtherItem(-1)
            ])));
        //
        // pointers
        $.getPointerPos = $.function(({ rect  })=>(e)=>new Point(e.pageX, e.pageY).translateSelf(rect.negate()));
        $.effect(({ host , minimap , pointers , getPointerPos , selection  })=>{
            const clear = $.atomic((e)=>{
                pointers.delete(e.pointerId);
                //!? 'pointers:', pointers.size
                if ($.state.is(SurfaceState.CenteringItem) || $.state.is(SurfaceState.CenteringView)) return;
                if (pointers.size === 0) {
                    requestAnimationFrame(()=>{
                        $.cursorState.returnToIdle();
                    });
                    if ($.state.is(SurfaceState.Selecting)) {
                        const rect = $.selectingRect.normalize($.viewMatrix);
                        const items = $.items.filter((x)=>x.rect.intersectsRect(rect));
                        $.state.emit(SurfaceState.Selecting, {
                            event: e,
                            items,
                            rect: $.selectingRect
                        });
                    } else {
                        $.transition.to(SurfaceState.Idle);
                    }
                } else if (pointers.size === 1) {
                    $.transition.to(SurfaceState.Panning);
                    $.pinchStartMatrix = null;
                }
            });
            return $.chain($.on(window).pointermove.throttle(16.66666)((e)=>{
                $.pointer = {
                    id: e.pointerId,
                    pos: getPointerPos(e)
                };
            }), $.on(host).pointerdown((e)=>{
                if (!(e.buttons & $.MouseButton.Left)) return;
                const pos = getPointerPos(e);
                pointers.set(e.pointerId, pos);
                //!? 'pointers:', pointers.size
                // if (!$.isMobile && ($.mode === SurfaceMode.CenteringItem || $.mode === SurfaceMode.CenteringView)) return
                if (pointers.size === 1) {
                    if (e.ctrlKey) {
                        $.selectingStartPos = pos.clone();
                        Object.assign(selection.style, new Rect(...$.selectingStartPos, 1, 1).toStyle());
                        $.transition.to(SurfaceState.Selecting);
                    } else if (e.composedPath().includes(minimap)) {
                        $.transition.to(SurfaceState.MinimapPanning);
                    } else {
                        $.transition.to(SurfaceState.Panning);
                    }
                } else if (pointers.size === 2) {
                    $.transition.to.expireAfter(200)(SurfaceState.Pinching);
                }
            }), $.on(window).pointercancel(clear), $.on(window).pointerup(clear));
        });
        $.effect(({ pointers  })=>$.on(window).pointermove($.with(({ viewMatrix , pointer  })=>$.when($.state.is(SurfaceState.Panning), (e)=>{
                    const id = e.pointerId;
                    if (pointer?.id !== id) return;
                    const p = pointers.get(id);
                    if (!p) return;
                    const n = pointer.pos;
                    pointers.set(id, n);
                    const d = n.screen(p).normalize(viewMatrix.a);
                    if (d.absoluteSum() === 0) return;
                    // $.transition.to(SurfaceState.Panning, () => {
                    $.viewMatrix = viewMatrix.translate(d.x, d.y);
                }))));
        // minimap panning
        $.effect(({ minimap , rect , viewRect , pointers , pixelRatio  })=>$.on(window).pointermove($.with(({ viewMatrix , pointer  })=>$.when($.state.is(SurfaceState.MinimapPanning), (e)=>{
                    const id = e.pointerId;
                    if (pointer?.id !== id) return;
                    const p = pointers.get(id);
                    if (!p) return;
                    const n = pointer.pos;
                    pointers.set(id, n);
                    const m = Point.fromObject(minimap.canvas).normalizeSelf(rect.size);
                    const s = viewRect.normalize(rect);
                    const scale = new Point(Math.max(0.01, s.width / m.width * pixelRatio), Math.max(0.01, s.height / m.height * pixelRatio));
                    const d = n.screen(p).scale(scale.negate());
                    if (d.absoluteSum() === 0) return;
                    // $.transition.to(SurfaceState.MinimapPanning, () => {
                    $.viewMatrix = viewMatrix.translate(d.x, d.y);
                }))));
        // pinching
        $.effect(({ pointers , getPointerPos  })=>$.on(window).pointermove($.callback(({ viewMatrix , pointer , minZoom , maxZoom , pinchStartDistance , pinchStartMatrix ,  })=>$.when($.state.is(SurfaceState.Pinching) && pointer, (e)=>{
                    const id = e.pointerId;
                    const p = pointers.get(id);
                    const n = getPointerPos(e);
                    const d = n.screen(p).normalize(viewMatrix.a);
                    pointers.set(id, n);
                    const [a, b] = [
                        ...pointers.values()
                    ];
                    if (!a || !b) return;
                    const dist = a.distance(b);
                    if (!dist) return;
                    if (!pinchStartMatrix) {
                        $.pinchStartDistance = dist;
                        $.pinchStartMatrix = viewMatrix.translate();
                    } else {
                        const scaleDiff = dist / pinchStartDistance;
                        const newScale = pinchStartMatrix.a * scaleDiff;
                        if (newScale < minZoom || newScale > maxZoom) return;
                        const c = a.translate(b).scaleSelf(0.5).normalizeSelf(viewMatrix);
                        // $.transition.to.expireAfter(200)(
                        //   $.atomic(() => {
                        $.viewMatrix = viewMatrix.translate().translateSelf(d.x, d.y).translateSelf(c.x, c.y);
                        $.viewMatrix.a = $.viewMatrix.d = newScale;
                        $.viewMatrix.translateSelf(-c.x, -c.y);
                    // )
                    }
                }))));
        // wheel zoom
        $.effect(({ minimap , minZoom , maxZoom , getPointerPos , viewFrameRect  })=>$.chain($.on(window).wheel.stop.prevent.not.passive(), $.on(window).wheel.passive.task($.with(({ viewMatrix  })=>$.when(!$.state.is(SurfaceState.CenteringItem) && !$.state.is(SurfaceState.CenteringView), (e)=>{
                    const mul = e.deltaMode === 1 ? 15 : 1.12;
                    const sign = Math.sign(e.deltaY);
                    const abs = Math.abs(e.deltaY) * mul;
                    const { a  } = viewMatrix;
                    let z = (1 - abs * sign * 0.0005 * a ** 0.001) ** 2.08;
                    if ((a > minZoom || z > 1) && (a < maxZoom || z < 1)) {
                        if (a * z < minZoom) z = minZoom / a;
                        if (a * z > maxZoom) z = maxZoom / a;
                        if (!isFinite(z)) return;
                        const screenPos = getPointerPos(e);
                        const pos = (e.composedPath().includes(minimap) ? viewFrameRect.size.scale(0.5) : screenPos).normalize(viewMatrix);
                        const paddedRect = viewFrameRect.zoomLinear(-150);
                        const item = $.hoveringItem;
                        // items.find(x =>
                        //   !x.points && pos.withinRect(x.rect.scaleLinear(0, 150 / a).translateSelf(0, -150 / a))
                        if (item) {
                            const rect = item.rect //.scaleLinear(0, 150 / a).translateSelf(0, -150 / a)
                            ;
                            const maxA = Math.min(paddedRect.height / rect.height, paddedRect.width / rect.width);
                            if (a * z > maxA) z = maxA / a;
                            let its = 100;
                            // eslint-disable-next-line no-constant-condition
                            while(true){
                                if (!--its) break;
                                const destRect = rect.transform(viewMatrix.translate(pos.x, pos.y).scaleSelf(z).translateSelf(-pos.x, -pos.y));
                                let moved = false;
                                const box = screenPos.withinRect(paddedRect) ? paddedRect : viewFrameRect;
                                if (destRect.top < box.top) {
                                    moved = true;
                                    viewMatrix.translateSelf(0, Math.abs(box.top - destRect.top));
                                } else if (destRect.bottom > box.bottom) {
                                    moved = true;
                                    viewMatrix.translateSelf(0, -Math.abs(destRect.bottom - box.bottom));
                                }
                                if (destRect.left < box.left) {
                                    moved = true;
                                    viewMatrix.translateSelf(Math.abs(box.left - destRect.left), 0);
                                } else if (destRect.right > box.right) {
                                    moved = true;
                                    viewMatrix.translateSelf(-Math.abs(destRect.right - box.right), 0);
                                }
                                if (!moved) break;
                            }
                        }
                        $.transition.to(SurfaceState.Wheeling, ()=>{
                            $.viewMatrix = viewMatrix.translate(pos.x, pos.y).scaleSelf(z).translateSelf(-pos.x, -pos.y);
                        });
                    }
                })))));
        // selecting
        $.effect(({ pointers , selection  })=>$.on(window).pointermove($.with(({ pointer , selectingStartPos: startPos  })=>$.when($.state.is(SurfaceState.Selecting), $.queue.raf((e)=>{
                    if (!startPos) return;
                    const id = e.pointerId;
                    if (pointer?.id !== id) return;
                    const p = pointers.get(id);
                    if (!p) return;
                    const n = pointer.pos;
                    pointers.set(id, n);
                    const endPos = n;
                    const left = Math.min(startPos.x, endPos.x);
                    const top = Math.min(startPos.y, endPos.y);
                    $.selectingRect = new Rect(left, top, Math.abs(endPos.x - startPos.x), Math.abs(endPos.y - startPos.y));
                    Object.assign(selection.style, $.selectingRect.toStyle());
                })))));
        // animation
        // we convert to string to prevent effect triggering unnecessary new animations
        $.viewMatrixString = $.reduce(({ viewMatrix  })=>viewMatrix.toString());
        if ($.isMobile) {
            $.effect.throttle(SurfaceAnimSettings[SurfaceState.Pinching].duration * 0.5).first.last.next(({ setViewStyleTransform , viewMatrixString  })=>{
                if (!$.state.is(SurfaceState.Pinching)) return;
                setViewStyleTransform(viewMatrixString);
            });
            $.effect.throttle(SurfaceAnimSettings[SurfaceState.Panning].duration).first.last.next(({ setViewStyleTransform , viewMatrixString  })=>{
                if (!$.state.is(SurfaceState.Panning)) return;
                setViewStyleTransform(viewMatrixString);
            });
            $.effect.raf(({ view , viewMatrixString  })=>{
                if ($.state.is(SurfaceState.Pinching) || $.state.is(SurfaceState.Panning)) return;
                view.style.transform = viewMatrixString;
            });
        } else {
            // panning
            $.effect.throttle(75).first.last.next(({ setViewStyleTransform , viewMatrixString  })=>{
                if (!$.state.is(SurfaceState.MinimapPanning) && !$.state.is(SurfaceState.Panning)) return;
                setViewStyleTransform(viewMatrixString);
            });
            // wheeling
            $.effect.throttle(75).last.next(({ setViewStyleTransform , viewMatrixString  })=>{
                if (!$.state.is(SurfaceState.Wheeling)) return;
                setViewStyleTransform(viewMatrixString);
            });
            $.effect.throttle(32).first.last.next(({ setViewStyleTransform , viewMatrixString  })=>{
                if ($.state.is(SurfaceState.MinimapPanning) || $.state.is(SurfaceState.Panning) || $.state.is(SurfaceState.Wheeling)) return;
                setViewStyleTransform(viewMatrixString);
            });
        }
        $.setViewStyleTransform = $.reduce(({ view  })=>$.queue.raf((matrixString)=>{
                //!timeLog 'animating'
                $.viewStyleTransform = matrixString;
                view.style.transform = matrixString;
            }));
        //
        // parts
        const MinimapPart = $.part(({ Minimap , rect , viewMatrix: matrix , items , pointers , pixelRatio , minimapScale , minimapRatio ,  })=>/*#__PURE__*/ jsx(Minimap, {
                ref: $.ref.minimap,
                part: "minimap",
                rect,
                matrix,
                items,
                pointers,
                pixelRatio,
                scale: minimapScale,
                ratio: minimapRatio
            }));
        // const GridPart = part((
        //   { Grid, rect, viewMatrix: matrix, animSettings, pixelRatio, gridCellSize: cellSize, xPattern, yPattern },
        $.render(()=>/*#__PURE__*/ jsxs(Fragment, {
                children: [
                    /*#__PURE__*/ jsx("style", {
                        children: style$1
                    }),
                    /*#__PURE__*/ jsx("div", {
                        ref: $.ref.viewFrame,
                        part: "frame",
                        children: /*#__PURE__*/ jsx("slot", {
                            ref: $.ref.view,
                            part: "view"
                        })
                    }),
                    /*#__PURE__*/ jsx(MinimapPart, {}),
                    /*#__PURE__*/ jsx("div", {
                        ref: $.ref.selection,
                        part: "selection"
                    }),
                    /*#__PURE__*/ jsx("div", {
                        part: "overlay"
                    }),
                    /*#__PURE__*/ jsx("div", {
                        part: "full-size",
                        ref: $.ref.fullSize
                    })
                ]
            }));
    }
};
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "state", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "cursorState", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "transition", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "pixelRatio", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "gridCellSize", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "snapThreshold", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "minZoom", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "maxZoom", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "minimapScale", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "minimapRatio", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "xPattern", void 0);
__decorate$3([
    $.attr()
], SurfaceElement.prototype, "yPattern", void 0);
SurfaceElement = __decorate$3([
    $.element()
], SurfaceElement);

var __decorate$2 = undefined && undefined.__decorate || function(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for(var i = decorators.length - 1; i >= 0; i--)if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
let SurfaceItemElement = class SurfaceItemElement extends HTMLElement {
    rect = new Rect();
    pos = $(this).reduce(({ $ , rect  })=>$.pos?.equals(rect.pos) ? $.pos : rect.pos);
    size = $(this).reduce(({ $ , rect  })=>$.size?.equals(rect.size) ? $.size : rect.size);
    surface;
    points;
    mounted($) {
        $.effect(({ host , pos  })=>{
            Object.assign(host.style, pos.toStylePosition());
        });
        $.effect(({ host , size  })=>{
            Object.assign(host.style, size.toStyleSize());
        });
    }
};
__decorate$2([
    $.out()
], SurfaceItemElement.prototype, "rect", void 0);
SurfaceItemElement = __decorate$2([
    $.element()
], SurfaceItemElement);

var __decorate$1 = undefined && undefined.__decorate || function(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for(var i = decorators.length - 1; i >= 0; i--)if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
const SurfaceMoveState = {
    Idle: 'surfacemoveidle',
    ItemMove: 'surfacemoveitemmove'
};
let SurfaceMoveElement = class SurfaceMoveElement extends HTMLElement {
    root = $(this).shadow( `<style>${`:host {
    display: inline-flex;
    width: 100%;
    height: 100%;
  }`}</style><slot></slot>`);
    state = $(this).state(SurfaceMoveState);
    surface;
    dest;
    pointerId;
    grabPos;
    mounted($) {
        $.effect(({ host  })=>{
            if (!~host.tabIndex) host.tabIndex = 0;
        });
        $.effect(({ state , surface  })=>{
            if (state.isIdle) {
                if (surface.state.is(SurfaceState.Overlay)) {
                    surface.state.pop(SurfaceState.Overlay);
                }
            } else {
                surface.state.push(SurfaceState.Overlay);
                surface.cursorState.push(SurfaceCursorState.Grabbing);
            }
        });
        $.effect(({ host , surface , dest  })=>$.ondblclick(host, ()=>{
                surface.centerItem(dest);
            }));
        $.effect(({ host , surface , dest  })=>$.on(host).pointerdown.stop($.when($.state.isIdle, $.atomic((e)=>{
                if (!(e.buttons & $.MouseButton.Left)) return;
                const pos = new Point(e.pageX, e.pageY).translateSelf(surface.rect.negate()).normalizeSelf(surface.viewMatrix);
                const rect = dest.rect //Rect.fromElement(dest)
                ;
                $.pointerId = e.pointerId;
                $.grabPos = rect.pos.translateSelf(pos.negateSelf())// this creates the effect of lifting up
                .translateSelf(-(surface.gridCellSize * surface.snapThreshold));
                // move the window over everything, also 1 bit above its cables
                dest.style.zIndex = (++SurfaceElement.zIndex << 1).toString();
                $.state.push(SurfaceMoveState.ItemMove, {
                    dest,
                    rect
                });
            }))));
        $.effect(({ surface , dest , grabPos , pointerId  })=>{
            if (!$.state.is(SurfaceMoveState.ItemMove)) return;
            const rect = dest.rect
            ;
            const snapPoint = (a, snapThreshold = surface.snapThreshold)=>{
                const snap = 1 / surface.gridCellSize;
                const b = a.translate(rect.size);
                const sa = a.precisionRound(snap);
                const sb = b.precisionRound(snap);
                const { xPattern: xp , yPattern: yp  } = surface;
                const ga = sa.pos.normalizeSelf(surface.gridCellSize);
                const gb = sb.pos.normalizeSelf(surface.gridCellSize);
                const iax = +xp[modwrap(ga.x, xp.length)];
                const iay = +yp[modwrap(ga.y, yp.length)];
                const ibx = +xp[modwrap(gb.x, xp.length)];
                const iby = +yp[modwrap(gb.y, yp.length)];
                const snapSize = surface.gridCellSize * snapThreshold;
                rect.x = iax && Math.abs(sa.x - a.x) < snapSize ? sa.x : ibx && Math.abs(sb.x - b.x) < snapSize ? sb.x - rect.width : a.x;
                rect.y = iay && Math.abs(sa.y - a.y) < snapSize ? sa.y : iby && Math.abs(sb.y - b.y) < snapSize ? sb.y - rect.height : a.y;
                dest.rect = rect.clone();
            };
            const offMove = $.on(window).pointermove((e)=>{
                if (e.pointerId !== pointerId) return;
                const pos = new Point(e.pageX, e.pageY).translateSelf(surface.rect.negate()).normalizeSelf(surface.viewMatrix).translateSelf(grabPos);
                snapPoint(pos);
                $.state.emit(SurfaceMoveState.ItemMove, {
                    dest,
                    rect
                });
            });
            const offUp = $.on(window).pointerup.once(()=>{
                snapPoint(rect.pos, 1);
                offMove();
                $.state.pop(SurfaceMoveState.ItemMove, {
                    dest,
                    rect
                });
            });
            // this creates the initial effect of lifting up
            snapPoint(new Point(rect.pos).translate(-(surface.gridCellSize * surface.snapThreshold)));
            return $.chain(offMove, offUp);
        });
    }
};
__decorate$1([
    $.attr()
], SurfaceMoveElement.prototype, "state", void 0);
SurfaceMoveElement = __decorate$1([
    $.element()
], SurfaceMoveElement);

var __decorate = undefined && undefined.__decorate || function(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for(var i = decorators.length - 1; i >= 0; i--)if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
const SurfaceResizeState = {
    Idle: 'surfaceresizeidle',
    ItemResize: 'surfaceresizeitemresize'
};
let SurfaceResizeElement = class SurfaceResizeElement extends HTMLElement {
    root = $(this).shadow(/*html*/ `<style>${`:host {
    display: inline-flex;
    width: 100%;
    height: 100%;
  }`}</style><slot></slot>`);
    state = $(this).state(SurfaceResizeState);
    surface;
    dest;
    pointerId;
    grabPos;
    orig;
    mounted($) {
        $.effect(({ host  })=>{
            if (!~host.tabIndex) host.tabIndex = 0;
        });
        $.effect(({ state , surface  })=>{
            if (state.isIdle) {
                if (surface.state.is(SurfaceState.Overlay)) {
                    surface.state.pop(SurfaceState.Overlay);
                }
            } else {
                surface.state.push(SurfaceState.Overlay);
                surface.cursorState.push(SurfaceCursorState.NWSEResize);
            }
        });
        $.effect(({ host , surface , dest  })=>$.on(host).pointerdown.stop($.when($.state.isIdle, $.atomic((e)=>{
                if (!(e.buttons & $.MouseButton.Left)) return;
                const pos = new Point(e.pageX, e.pageY).translateSelf(surface.rect.negate()).normalizeSelf(surface.viewMatrix);
                $.pointerId = e.pointerId;
                $.grabPos = pos.negateSelf();
                $.orig = dest.rect.clone();
                // move the window over everything, also 1 bit above its cables
                dest.style.zIndex = (++SurfaceElement.zIndex << 1).toString();
                $.state.push(SurfaceResizeState.ItemResize, {
                    dest,
                    rect: $.orig
                });
            }))));
        $.effect(({ surface , dest , grabPos , pointerId , orig  })=>{
            if (!$.state.is(SurfaceResizeState.ItemResize)) return;
            const rect = dest.rect;
            let hasMoved = false;
            const snapPoint = (p, snapThreshold = surface.snapThreshold)=>{
                hasMoved = true;
                const snap = 1 / surface.gridCellSize;
                const sp = round(p, snap);
                const snapSize = surface.gridCellSize * snapThreshold;
                const { xPattern , yPattern  } = surface;
                const ipx = +xPattern[modwrap(sp.x / surface.gridCellSize, xPattern.length)];
                const ipy = +yPattern[modwrap(sp.y / surface.gridCellSize, yPattern.length)];
                const x = ipx && Math.abs(sp.x - p.x) < snapSize ? sp.x : p.x;
                const y = ipy && Math.abs(sp.y - p.y) < snapSize ? sp.y : p.y;
                rect.width = Math.max(10, x - rect.left);
                rect.height = Math.max(10, y - rect.top);
                dest.rect = rect.clone();
            };
            const offMove = $.on(window).pointermove((e)=>{
                if (e.pointerId !== pointerId) return;
                const pos = new Point(e.pageX, e.pageY).translateSelf(surface.rect.negate()).normalizeSelf(surface.viewMatrix).translateSelf(grabPos).translateSelf(orig.size).translateSelf(rect);
                snapPoint(pos);
                $.state.emit(SurfaceResizeState.ItemResize, {
                    dest,
                    rect
                });
            });
            const offUp = $.on(window).pointerup.once(()=>{
                if (hasMoved) snapPoint(rect.size.translate(rect.pos), 1);
                offMove();
                $.state.pop(SurfaceResizeState.ItemResize, {
                    dest,
                    rect
                });
            });
            return $.chain(offMove, offUp);
        });
    }
};
__decorate([
    $.attr()
], SurfaceResizeElement.prototype, "state", void 0);
SurfaceResizeElement = __decorate([
    $.element()
], SurfaceResizeElement);

class Popup {
    static create(data) {
        const popup = new Popup(data);
        popup.attach();
        popup.create();
        return popup;
    }
    id = cheapRandomId();
    scene;
    mode = 'local';
    vel = new Point();
    rect;
    rectDest = new Rect();
    prevPos = new Point();
    placement;
    originalPlacement;
    center;
    contentsRect;
    destRect;
    $;
    context;
    viewMatrix;
    // physics
    place;
    collisions = new Map();
    viewportIntersection = Intersect.None;
    exceedsViewport = false;
    destExceedsViewport = false;
    destWithinViewport = false;
    surface;
    constructor(data = {}) {
        Object.assign(this, data);
    }
    attach() {
        $.ContextClass.attach(this);
    }
    create() {
        const { $ } = this;
        $.place = $.reduce(({ contentsRect, destRect }) => placement => contentsRect.place(destRect, placement));
        $.surface = $.fulfill(({ scene }) => fulfill => scene.$.effect(({ surface }) => fulfill(surface)));
        $.effect.raf(({ surface, scene, destRect: _ }) => {
            if (surface.state.is(SurfaceState.Overlay)) {
                scene?.runCollisions?.();
            }
        });
    }
}

const style = /*css*/ `
${Object.entries(SurfaceAnimSettings).map(([name, { duration, easing }]) => /*css*/ `
    :host([placed][transition=${name}]:not([rigid])) {
      transition:
        transform ${duration}ms cubic-bezier(${easing}),
        opacity 200ms ease-out;
    }
  `).join('')}
:host {
  position: absolute;
  transform: matrix(1,0,0,1,0,0);
  -webkit-tap-highlight-color: transparent;
}
/* :host([rigid]), */
:host(:not(:focus)) {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -khtml-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  outline: none !important;
}
/* :host:focus {
  outline: none !important;
} */
:host(:not([placed])) {
  opacity: 0;
}
/* :host(:not([placed]):not([rigid])) {
  transition: opacity 200ms ease-out;
} */
/* :host([placed]:not([rigid])) {
  transition:
    transform 150ms ease-out,
    opacity 200ms ease-out;
} */
:host([placed][rigid]) {
  transition:
    transform 150ms ease-out;
}
:host(:not([placed]):not([rigid])) [part=contents] {
  transform: scale(0.1, 0.9) translate(-15px, 7px);
}
[part=contents] {
  /* transform-origin: 10% 50%;
  transition: transform 350ms cubic-bezier(0.55,0,0.4,1.0); */
  box-sizing: border-box;
  display: inline-flex;
}
:host(:not([rigid])) [part=contents] {
  transform-origin: 10% 60%;
  transition: transform 300ms cubic-bezier(0.15,0,0.05,1.0);
}
`;
let PopupElement = class PopupElement extends HTMLElement {
    placement = 'n';
    placed = false;
    center = false;
    rigid = false;
    transition = SurfaceState.Idle;
    surface;
    scene;
    contents;
    contentsRect;
    dest;
    destRect;
    popup;
    rect;
    pos;
    viewMatrix;
    viewportRect;
    hostMatrixString;
    setHostStyleTransform;
    mounted($) {
        $.effect(({ host }) => {
            host.tabIndex = 0;
        });
        // place popup in scene
        $.effect(({ popup, scene }) => {
            scene.popups.add(popup);
            popup.attach();
            popup.create();
            $.scene = scene;
            popup.scene = scene;
            if ($.rigid) {
                scene.runCollisions?.();
            }
            return () => scene.popups.delete(popup);
        });
        // read data from popup
        $.rect = $.fulfill(({ popup }) => (fulfill => popup.$.effect(({ rect }) => fulfill(rect))));
        $.pos = $.reduce(({ rect }) => rect.pos);
        // write data to popup
        $.popup = $.reduce(({ viewMatrix, center, destRect, contentsRect, placement }) => {
            const data = {
                center,
                contentsRect,
                // placement is set once from the element and then
                // handled entirely in the worker
                // placement: placement,
                originalPlacement: placement,
                // we move the dest rectangle to our (normal dom) screen space
                // because its dimensions are in surface space (zoomed/panned etc)
                destRect: destRect.transform(viewMatrix),
            };
            //!? 'data', data
            if ($.popup) {
                return Object.assign($.popup, data);
            }
            else
                return Popup.create(data);
        });
        // read contents
        $.contentsRect = $.fulfill(({ contents }) => (fulfill => (fulfill(Rect.fromElement(contents)), //
            $.observe.resize(contents, () => fulfill(Rect.fromElement(contents))))));
        // read dest
        $.destRect = $.fulfill(({ dest }) => (fulfill => dest.$.effect(({ rect }) => fulfill(rect))));
        // read surface
        $.effect(({ surface }) => surface.$.effect(({ transition }) => {
            //!? 'transition', transition.current
            $.transition = transition.current;
        }));
        // read scene
        $.effect(({ scene }) => scene.$.effect(({ viewMatrix, viewportRect }) => {
            $.viewMatrix = viewMatrix;
            $.viewportRect = viewportRect;
        }));
        // draw
        $.hostMatrixString = $.reduce(({ rect, scene }) => {
            const p = rect.contain(scene.viewportRect);
            return new Matrix()
                .translateSelf(p.x, p.y)
                .toString();
        });
        $.effect(({ host, pos }) => {
            const transformString = `translate(${pos.x}px,${pos.y}px)`;
            host.style.transform = transformString;
            if (!host.placed) {
                setTimeout(() => {
                    host.placed = true;
                }, 100);
            }
        });
        $.render(() => (jsxs(Fragment, { children: [jsx("style", { children: style }), jsx("slot", { ref: $.ref.contents, part: "contents" })] })));
    }
};
__decorate$6([
    $.attr(),
    __metadata("design:type", String)
], PopupElement.prototype, "placement", void 0);
__decorate$6([
    $.attr(),
    __metadata("design:type", Object)
], PopupElement.prototype, "placed", void 0);
__decorate$6([
    $.attr(),
    __metadata("design:type", Object)
], PopupElement.prototype, "center", void 0);
__decorate$6([
    $.attr(),
    __metadata("design:type", Object)
], PopupElement.prototype, "rigid", void 0);
__decorate$6([
    $.attr(),
    __metadata("design:type", Object)
], PopupElement.prototype, "transition", void 0);
PopupElement = __decorate$6([
    $.element()
], PopupElement);

export { PopupElement, PopupScene };
//# sourceMappingURL=x-popup.js.map
